"""Contains the basic definition of an observed spectrum for TauRex 3."""

import typing as t

import numpy as np
import numpy.typing as npt

from taurex.binning import Binner
from taurex.core import DerivedType, Fittable, FittingType
from taurex.log import Logger
from taurex.output import OutputGroup
from taurex.output.writeable import Writeable


class BaseSpectrum(Logger, Fittable, Writeable):
    """A base class that represents spectrums.

    *Abstract class*

    A base class where spectrums are loaded (or later created). This
    is used to either plot against the forward model or passed into the
    optimizer to be used to fit the forward model.
    """

    def __init__(self, name: str):
        """Initialize.

        Parameters
        ----------

        name : str
            Name to be used in logging

        """
        Logger.__init__(self, name)
        Fittable.__init__(self)

    def create_binner(self) -> Binner:
        """Creates the appropriate binning object."""
        from taurex.binning import FluxBinner

        return FluxBinner(wngrid=self.wavenumberGrid, wngrid_width=self.binWidths)

    @property
    def spectrum(self) -> npt.NDArray[np.float64]:
        """Spectrum of the observation.

        **Requires Implementation**


        Should return the observed spectrum.

        Raises
        ------
        NotImplementedError

        """
        raise NotImplementedError

    @property
    def rawData(self) -> t.Any:  # noqa: N802
        """Raw data of the observation.
        **Requires Implementation**


        Should return the raw data set.

        Raises
        ------
        NotImplementedError

        """
        raise NotImplementedError

    @property
    def wavelengthGrid(self) -> npt.NDArray[np.float64]:  # noqa: N802
        """Wavelength grid of the spectrum in microns.


        **Requires Implementation**


        Should return the wavelength grid of the spectrum in microns.
        This does not need to necessarily match the shape of :func:`spectrum`

        Raises
        ------
        NotImplementedError

        """
        raise NotImplementedError

    @property
    def wavenumberGrid(self):  # noqa: N802
        """Wavenumber grid in :math:`cm^{-1}`

        Returns
        -------
        wngrid : :obj:`array`

        """
        return 10000 / self.wavelengthGrid

    @property
    def binEdges(self) -> npt.NDArray[np.float64]:  # noqa: N802
        """Bin edges of the wavenumber grid.
        **Requires Implementation**


        Should return the bin edges of the wavenumber grid

        Raises
        ------
        NotImplementedError

        """
        raise NotImplementedError

    @property
    def binWidths(self) -> npt.NDArray[np.float64]:  # noqa: N802
        """Widths of each bin in the wavenumber grid
        **Requires Implementation**


        Should return the widths of each bin in the wavenumber grid

        Raises
        ------
        NotImplementedError

        """
        raise NotImplementedError

    @property
    def errorBar(self) -> npt.NDArray[np.float64]:  # noqa: N802
        """Return error or uncertainty of the spectrum.
        **Requires Implementation**


        Should return the error. *Must* be the same shape as
        :func:`spectrum`

        Raises
        ------
        NotImplementedError

        """
        raise NotImplementedError

    @property
    def fittingParameters(self) -> t.Dict[str, FittingType]:  # noqa: N802
        """Return fitting parameters."""
        return self.fitting_parameters()

    @property
    def derivedParameters(self) -> t.Dict[str, DerivedType]:  # noqa: N802
        """Return derived parameters."""
        return self.derived_parameters()

    def write(self, output: OutputGroup) -> OutputGroup:
        """Write spectrum to output group."""
        output.write_array("wlgrid", self.wavelengthGrid)
        output.write_array("spectrum", self.spectrum)
        output.write_array("binedges", self.binEdges)
        output.write_array("binwidths", self.binWidths)
        output.write_array("errorbars", self.errorBar)

        return output

    @classmethod
    def input_keywords(cls) -> t.Tuple[str, ...]:
        raise NotImplementedError



import typing as t
from functools import partial

from taurex.chemistry import Chemistry
from taurex.contributions import Contribution
from taurex.core.priors import Prior
from taurex.instruments import Instrument
from taurex.log import setup_log
from taurex.model import ForwardModel
from taurex.optimizer import Optimizer
from taurex.planet import Planet
from taurex.pressure import PressureProfile
from taurex.spectrum import BaseSpectrum
from taurex.stellar import Star
from taurex.temperature import TemperatureProfile
from taurex.types import PathLike

from .classfactory import ClassFactory, DiscoverMethod, InputKeywordMethod

log = setup_log(__name__)

T = t.TypeVar("T")

InputT = t.TypeVar("InputT", bound=InputKeywordMethod)
DiscoverT = t.TypeVar("DiscoverT", bound=DiscoverMethod)


class ConfigType(t.TypedDict):
    """Input section."""

    type: t.Optional[str]
    profile_type: t.Optional[str]
    python_file: t.Optional[str]
    ...


def get_keywordarg_dict(klass: t.Type[T]) -> t.Tuple[t.Dict[str, t.Any], bool]:
    """Get all keyword arguments for a class."""
    import inspect

    from taurex.mixin.core import Mixin

    is_mixin = issubclass(klass, Mixin)
    has_kvar = False
    init_dicts = {}
    if not is_mixin:
        init_dicts = {}
        signature = inspect.signature(klass.__init__)

        parameters = signature.parameters

        args = list(parameters.keys())

        if "self" in args:
            args.remove("self")

        for arg in args:
            if parameters[arg].kind == inspect.Parameter.VAR_KEYWORD:
                has_kvar = True
                continue
            if parameters[arg].default == inspect.Parameter.empty:
                init_dicts[arg] = None
            else:
                init_dicts[arg] = parameters[arg].default
    else:
        from taurex.mixin.core import determine_mixin_args

        init_dicts, has_kvar = determine_mixin_args(klass.__bases__)

    log.debug("Init dicts are %s", init_dicts)
    log.debug("Has kvar %s", has_kvar)
    return init_dicts, has_kvar


def create_klass(config: ConfigType, klass: t.Type[T]) -> T:
    """Create a class from a dictionary."""
    kwargs, has_kvar = get_keywordarg_dict(klass)

    for key in config:
        if key in kwargs or has_kvar:
            value = config[key]
            kwargs[key] = value
        else:
            log.error(f"Object {klass.__name__} does not have parameter {key}")
            log.error("Available parameters are %s", kwargs.keys())
            raise KeyError(f"Object {klass.__name__} does not have parameter {key}")
    obj = klass(**kwargs)
    return obj


def generic_factory(profile_type: str, baseclass: t.Type[InputT]) -> t.Type[InputT]:
    """Get a class from input_keyword string."""
    cf = ClassFactory()
    return cf.find_klass_from_keyword(profile_type, baseclass)


# def create_profile(
#     config: ConfigType,
#     baseclass: t.Type[T],
#     alt_type: t.Optional[t.Union[str, t.Sequence[str]]] = None,
#     default_type: t.Optional[str] = None,
# ) -> T:
#     """Create a profile from a dictionary."""
#     config, klass, mixin = determine_klass(
#         config,
#         lambda x: generic_factory(x, baseclass),
#         baseclass=baseclass,
#         alt_type=alt_type,
#         default_type=default_type,
#     )


#     return obj


def create_chemistry(config: ConfigType) -> Chemistry:
    """Chemistry as a special case."""
    from taurex.chemistry import TaurexChemistry
    from taurex.data.profiles.chemistry.gas.gas import Gas

    gases = []
    gas_configs = {k: v for k, v in config.items() if isinstance(v, dict)}

    gases = [
        create_generic(
            {**v, "molecule_name": k},
            Gas,
            alt_type="gas_type",
        )
        for k, v in gas_configs.items()
    ]

    for k in gas_configs.keys():
        config.pop(k)

    log.debug(f"leftover keys {config}")

    obj = create_generic(config, Chemistry, alt_type="chemistry_type")

    if hasattr(obj, "addGas"):
        obj = t.cast(TaurexChemistry, obj)
        for g in gases:
            obj.addGas(g)

    return obj


def determine_klass(
    config: ConfigType,
    baseclass: t.Type[InputT],
    default_type: t.Optional[str] = None,
    alt_type: t.Optional[t.Union[str, t.Sequence[str]]] = None,
) -> t.Tuple[ConfigType, t.Type[InputT], bool]:
    """Determine class from input."""
    from taurex.mixin.core import build_new_mixed_class, find_mapped_mixin

    keyword_types = ["type", "profile_type"]
    if alt_type is not None:
        alt_type = [alt_type] if isinstance(alt_type, str) else alt_type
        keyword_types = keyword_types + alt_type

    check_no_type = len([k for k in keyword_types if k in config]) == 0

    if check_no_type and default_type is not None:
        config["type"] = default_type

    # Check if multiple keyword types exist
    # if they do then we should flag this as an error.

    avail_types = [k for k in keyword_types if k in config]

    if len(avail_types) > 1:
        log.error("Multiple type identifiers found in %s", config)
        raise ValueError("Multiple type identifiers found in %s", config)

    if len(avail_types) == 0:
        log.error("No type identifier found in %s", config)
        raise ValueError(
            f"No keyword types found in {config}, please "
            "include at least 'type' or "
            "'profile_type' in input",
        )

    field = avail_types[0]

    klass_field: t.Union[str, t.Sequence[str]] = config.pop(field)

    klass = None
    is_mixin = False
    if klass_field == "custom":
        try:
            python_file = config.pop("python_file")
        except KeyError as e:
            log.error("No python file for custom profile/model")
            raise KeyError from e

        return config, detect_and_return_klass(python_file, baseclass), False

    split = klass_field.split("+") if isinstance(klass_field, str) else klass_field
    if len(split) == 1:
        klass = generic_factory(klass_field, baseclass)
    else:
        is_mixin = True
        main_klass = generic_factory(split[-1], baseclass)

        mixin_class = find_mapped_mixin(main_klass)

        mixins = [generic_factory(s, mixin_class) for s in split[:-1]]
        klass = build_new_mixed_class(main_klass, mixins)

    return config, klass, is_mixin


def generate_contributions(config: ConfigType) -> t.List[Contribution]:
    """Generate contributions from input."""
    cf = ClassFactory()

    contribs_key = {k: v for k, v in config.items() if isinstance(v, dict)}

    contributions = []

    for key, value in contribs_key.items():
        klass = cf.find_klass_from_keyword(key, Contribution)
        contributions.append(create_klass(value, klass))

    return contributions


def create_generic(
    config: ConfigType,
    baseclass: t.Type[InputT],
    default_type: t.Optional[str] = None,
    alt_type: t.Optional[t.Union[str, t.Sequence[str]]] = None,
) -> InputT:
    """Create a generic class from input."""
    config, klass, _ = determine_klass(
        config,
        baseclass,
        default_type=default_type,
        alt_type=alt_type,
    )

    obj = create_klass(config, klass)

    return obj


def create_prior(prior: str) -> Prior:
    from taurex.util.fitting import parse_priors

    prior_name, args = parse_priors(prior)
    cf = ClassFactory()
    for p in cf.priorKlasses:
        if prior_name in (
            p.__name__,
            p.__name__.lower(),
            p.__name__.upper(),
        ):
            return p(**args)
    else:
        raise ValueError("Unknown Prior Type in input file")


def create_model(
    config: ConfigType,
    gas: Chemistry,
    temperature: TemperatureProfile,
    pressure: PressureProfile,
    planet: Planet,
    star: Star,
    observation: t.Optional[BaseSpectrum] = None,
) -> ForwardModel:
    """Create a forward model from input."""
    log.debug(config)
    config, klass, is_mixin = determine_klass(
        config, ForwardModel, alt_type="model_type"
    )

    log.debug(f"Chosen_model is {klass}")
    kwargs, has_kvar = get_keywordarg_dict(klass)
    log.debug(f"Model kwargs {kwargs}")
    log.debug(f"---------------{gas} {gas.activeGases}--------------")
    if "planet" in kwargs:
        kwargs["planet"] = planet
    if "star" in kwargs:
        kwargs["star"] = star
    if "chemistry" in kwargs:
        kwargs["chemistry"] = gas
    if "temperature_profile" in kwargs:
        kwargs["temperature_profile"] = temperature
    if "pressure_profile" in kwargs:
        kwargs["pressure_profile"] = pressure
    if "observation" in kwargs:
        kwargs["observation"] = observation
    log.debug(f"New Model kwargs {kwargs}")
    log.debug("Creating model---------------")

    kwargs.update({k: v for k, v in config.items() if not isinstance(v, dict)})
    obj = klass(**kwargs)

    contribs = generate_contributions(config)

    for c in contribs:
        obj.add_contribution(c)

    return obj


def detect_and_return_klass(
    python_file: PathLike, baseclass: t.Type[InputT]
) -> t.Type[InputT]:
    """Detect and return class from python file."""
    import importlib.util
    import inspect
    import pathlib

    python_file = pathlib.Path(python_file)

    if not python_file.exists():
        log.error("File %s does not exist", python_file)
        raise FileNotFoundError(f"File {python_file} does not exist")

    spec = importlib.util.spec_from_file_location("foo", python_file)
    foo = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(foo)
    classes = [
        m[1]
        for m in inspect.getmembers(foo, inspect.isclass)
        if m[1] is not baseclass and issubclass(m[1], baseclass)
    ]

    if len(classes) == 0:
        log.error("Could not find class of type %s in file %s", baseclass, python_file)
        raise Exception(f"No class inheriting from {baseclass} in " f"{python_file}")
    return classes[0]


# Wrap creation functions for convenience.
create_planet: t.Callable[[ConfigType], Planet] = partial(
    create_generic, baseclass=Planet, alt_type="planet_type", default_type="simple"
)
create_optimizer: t.Callable[[ConfigType], Optimizer] = partial(
    create_generic, baseclass=Optimizer, alt_type="optimizer"
)
create_instrument: t.Callable[[ConfigType], Instrument] = partial(
    create_generic, baseclass=Instrument, alt_type="instrument"
)
create_star: t.Callable[[ConfigType], Star] = partial(
    create_generic, baseclass=Star, alt_type="star_type"
)
create_temperature_profile: t.Callable[[ConfigType], TemperatureProfile] = partial(
    create_generic, baseclass=TemperatureProfile
)
create_pressure_profile: t.Callable[[ConfigType], PressureProfile] = partial(
    create_generic, baseclass=PressureProfile
)
create_spectrum: t.Callable[[ConfigType], BaseSpectrum] = partial(
    create_generic, baseclass=BaseSpectrum, alt_type="observation"
)
create_observation: t.Callable[[ConfigType], BaseSpectrum] = partial(
    create_generic, baseclass=BaseSpectrum, alt_type="observation"
)



"""Core mixin functions."""

import typing as t

from taurex.chemistry import Chemistry, Gas
from taurex.contributions import Contribution
from taurex.instruments import Instrument
from taurex.log import setup_log
from taurex.model import ForwardModel
from taurex.optimizer import Optimizer
from taurex.planet import Planet
from taurex.pressure import PressureProfile
from taurex.spectrum import BaseSpectrum
from taurex.stellar import Star
from taurex.temperature import TemperatureProfile

from ..core import Citable, Fittable

if t.TYPE_CHECKING:
    # Useful for type checking but not for runtime
    _BaseStar = Star
    _BaseTemperatureProfile = TemperatureProfile
    _BasePressureProfile = PressureProfile
    _BasePlanet = Planet
    _BaseContribution = Contribution
    _BaseChemistry = Chemistry
    _BaseForwardModel = ForwardModel
    _BaseSpectrum = BaseSpectrum
    _BaseOptimizer = Optimizer
    _BaseInstrument = Instrument
    _BaseGas = Gas

else:
    _BaseStar = object
    _BaseTemperatureProfile = object
    _BasePressureProfile = object
    _BasePlanet = object
    _BaseContribution = object
    _BaseChemistry = object
    _BaseForwardModel = object
    _BaseSpectrum = object
    _BaseOptimizer = object
    _BaseInstrument = object
    _BaseGas = object


# Try and create __init_mixin__

_log = setup_log(__name__)

T = t.TypeVar("T")

M = t.TypeVar("M", bound="Mixin")


class MixinProtocol(t.Protocol):
    """Mixin protocol."""

    def __init_mixin__(self, **kwargs: t.Dict[str, t.Any]) -> None:
        ...


def mixed_init(self, **kwargs: t.Dict[str, t.Any]) -> None:
    """Initialisation function for mixed classes."""
    import inspect

    new_class = self.__class__
    base_class = self.__class__.__bases__[-1]
    args = list(inspect.signature(base_class.__init__).parameters.keys())

    # Remove self
    if "self" in args:
        args.remove("self")
    new_kwargs = {}
    for k, v in kwargs.items():
        if k in args:
            new_kwargs[k] = v

    for klass in reversed(new_class.__bases__):
        klass.__init__(self, **new_kwargs)

    new_kwargs = {}

    for klass in reversed(new_class.__bases__[:-1]):
        klass = t.cast(t.Type[Mixin], klass)
        args = list(inspect.signature(klass.__init_mixin__).parameters.keys())
        if "self" in args:
            args.remove("self")
        new_kwargs = {}
        for k, v in kwargs.items():
            if k in args:
                new_kwargs[k] = v
        klass.__init_mixin__(self, **new_kwargs)
    # Embed class bases into object


class Mixin(MixinProtocol, Fittable, Citable, t.Generic[T]):
    """Base mixin class."""

    KLASS_COMPAT: t.Type[T] = None

    def __init__(self, **kwargs) -> None:
        """Constructor.

        Should not be called directly.

        """
        old_fitting_parameters = {}
        old_derived_parameters = {}
        if hasattr(self, "_param_dict"):
            old_fitting_parameters = self._param_dict
            old_derived_parameters = self._derived_dict
        super().__init__(**kwargs)

        if not hasattr(self, "_param_dict"):
            self._param_dict = {}
            self._derived_dict = {}

        self._param_dict.update(old_fitting_parameters)
        self._derived_dict.update(old_derived_parameters)

    def __init_mixin__(self):
        """Main initialisation function for mixin.

        This should be implemented by the mixin class and not ``__init__``.

        """
        pass

    @classmethod
    def input_keywords(cls) -> t.Tuple[str, ...]:
        raise NotImplementedError

    @classmethod
    def compatible(cls, other: t.Type) -> bool:
        if cls.KLASS_COMPAT:
            return issubclass(other, cls.KLASS_COMPAT)
        else:
            return False


class AnyMixin(Mixin[t.Any]):
    """Can enhance any class."""

    KLASS_COMPAT = object


class StarMixin(Mixin[Star], _BaseStar):
    """Enhances :class:`~taurex.data.stellar.star.Star`."""

    KLASS_COMPAT = Star


class TemperatureMixin(Mixin[TemperatureProfile], _BaseTemperatureProfile):
    """Enhances :class:`~taurex.data.profiles.temperature.TemperatureProfile`."""

    KLASS_COMPAT = TemperatureProfile


class PlanetMixin(Mixin[Planet], _BasePlanet):
    """Enhances :class:`~taurex.data.planet.Planet`."""

    KLASS_COMPAT = Planet


class ContributionMixin(Mixin[Contribution], _BaseContribution):
    """Enhances :class:`~taurex.contributions.Contribution`."""

    KLASS_COMPAT = Contribution


class ChemistryMixin(Mixin[Chemistry], _BaseChemistry):
    """Enhances :class:`~taurex.data.chemistry.Chemistry`."""

    KLASS_COMPAT = Chemistry


class PressureMixin(Mixin[PressureProfile], _BasePressureProfile):
    """Enhances :class:`~taurex.data.profiles.pressure.PressureProfile`."""

    KLASS_COMPAT = PressureProfile


class ForwardModelMixin(Mixin[ForwardModel], _BaseForwardModel):
    """Enhances :class:`~taurex.model.model.ForwardModel`."""

    KLASS_COMPAT = ForwardModel


class SpectrumMixin(Mixin[BaseSpectrum], _BaseSpectrum):
    """Enhances :class:`~taurex.spectrum.Spectrum`."""

    KLASS_COMPAT = BaseSpectrum


class ObservationMixin(SpectrumMixin):
    """Enhances :class:`~taurex.spectrum.Spectrum`."""

    pass


class OptimizerMixin(Mixin[Optimizer], _BaseOptimizer):
    """Enhances :class:`~taurex.optimizers.Optimizer`."""

    KLASS_COMPAT = Optimizer


class GasMixin(Mixin[Gas], _BaseGas):
    """Enhances :class:`~taurex.data.gas.Gas`."""

    KLASS_COMPAT = Gas


class InstrumentMixin(Mixin[Instrument], _BaseInstrument):
    """Enhances :class:`~taurex.instruments.instrument.Instrument`."""

    KLASS_COMPAT = Instrument


def determine_mixin_args(
    klasses: t.Sequence[t.Union[t.Type[T], t.Type[M]]]
) -> t.Tuple[t.Dict[str, t.Any], bool]:
    """Determine all arguments for a mixin class."""
    import inspect

    has_kvar = False
    all_kwargs = {}
    for klass in klasses:
        argspec = inspect.signature(klass.__init__)
        if issubclass(klass, Mixin):
            argspec = inspect.signature(klass.__init_mixin__)

        parameters = argspec.parameters
        args = list(parameters.keys())
        if "self" in args:
            args.remove("self")

        for arg in args:
            if parameters[arg].kind == inspect.Parameter.VAR_KEYWORD:
                has_kvar = True
                continue
            value = parameters[arg].default
            if value == inspect.Parameter.empty:
                all_kwargs[arg] = None
            else:
                all_kwargs[arg] = value

    _log.debug("All kwargs are %s", all_kwargs)
    return all_kwargs, has_kvar


def build_new_mixed_class(
    base_klass: t.Type[T], mixins: t.Sequence[t.Type[M]]
) -> t.Type[t.Union[T, M]]:
    """Build a new mixed class.

    Parameters
    ----------
    base_klass:
        Base class to mix with

    mixins:
        Sequence of mixin classes


    """
    if not hasattr(mixins, "__len__"):
        mixins = [mixins]

    all_classes = tuple(mixins) + tuple([base_klass])
    new_name = "+".join([x.__name__[:10] for x in all_classes])

    new_klass = type(new_name, all_classes, {"__init__": mixed_init})

    return new_klass


def enhance_class(
    base_klass: t.Type[T],
    mixins: t.Sequence[t.Type[M]],
    **kwargs: t.Dict[str, t.Any],
) -> t.Union[T, M]:
    """Build and initialise a new enhanced class.

    Parameters
    ----------
    base_klass:
        Base class to mix with

    mixins:
        Sequence of mixin classes

    kwargs:
        Keyword arguments to pass to the constructor

    Returns
    -------
    t.Union[T, M]:
        A new class that is a subclass of base_klass and all mixins

    Raises
    ------
    KeyError:
        If a keyword argument is not available in the new class

    """
    new_klass = build_new_mixed_class(base_klass, mixins)
    all_kwargs, has_kwarg = determine_mixin_args(new_klass.__bases__)

    for k in kwargs:
        if k not in all_kwargs and not has_kwarg:
            _log.error("Object %s does not have " "parameter %s", new_klass, k)
            _log.error("Available parameters are %s", all_kwargs)
            raise KeyError(f"Object {new_klass} does not have parameter {k}")
    return new_klass(**kwargs)


_mixin_map = {
    TemperatureProfile: TemperatureMixin,
    PressureProfile: PressureMixin,
    Planet: PlanetMixin,
    Star: StarMixin,
    Contribution: ContributionMixin,
    Chemistry: ChemistryMixin,
    ForwardModel: ForwardModelMixin,
    BaseSpectrum: SpectrumMixin,
    Optimizer: OptimizerMixin,
    Instrument: InstrumentMixin,
    Gas: GasMixin,
}
"""Map of base classes to mixin classes."""


def find_mapped_mixin(
    klass: t.Type[T],
) -> t.Type[Mixin]:
    """Find a mapped mixin class.

    Parameters
    ----------
    klass:
        Class to find the map.


    """
    for k in _mixin_map.keys():
        if issubclass(klass, k):
            return _mixin_map[k]

    raise ValueError(f"Class {klass} not found in mixin map")



"""Opacity for testing purposes."""
import typing as t

import numpy as np
import numpy.typing as npt

from taurex.util import create_grid_res

from . import InterpolatingOpacity


class FakeOpacity(InterpolatingOpacity):
    """Fake opacity for testing purposes."""

    def __init__(
        self,
        molecule_name: str,
        wn_res: t.Optional[float] = 15000,
        wn_size: t.Optional[t.Tuple[float, float]] = (300, 30000),
        num_p: t.Optional[int] = 20,
        num_t: t.Optional[int] = 27,
    ):
        super().__init__("FAKE")
        self._molecule_name = molecule_name
        self._wavenumber_grid = create_grid_res(wn_res, *wn_size)[:, 0]
        self._xsec_grid = np.random.rand(num_p, num_t, self._wavenumber_grid.shape[0])

        self._temperature_grid = np.linspace(100, 10000, num_t)
        self._pressure_grid = np.logspace(-6, 6, num_p)

    @property
    def moleculeName(self) -> str:  # noqa: N802
        """Name of molecule."""
        return self._molecule_name

    @property
    def xsecGrid(self) -> npt.NDArray[np.float64]:  # noqa: N802
        """Opacity grid."""
        return self._xsec_grid

    @property
    def wavenumberGrid(self) -> npt.NDArray[np.float64]:  # noqa: N802
        """Wavenumber grid."""
        return self._wavenumber_grid

    @property
    def temperatureGrid(self) -> npt.NDArray[np.float64]:  # noqa: N802
        """Temperature grid."""
        return self._temperature_grid

    @property
    def pressureGrid(self) -> npt.NDArray[np.float64]:  # noqa: N802
        """Pressure grid."""
        return self._pressure_grid



"""Lightcurve data class."""
import logging
import typing as t

import numpy as np
import numpy.typing as npt

from taurex.log import Logger
from taurex.output import OutputGroup, Writeable


class LightCurveInstrumentData(t.TypedDict):
    """Lightcurve data for a single instrument."""

    data: npt.NDArray[np.float64]
    time_series: npt.NDArray[np.float64]
    ld_coeff: npt.NDArray[np.float64]


InstrumentType = t.Literal["wfc3", "spitzer", "stis", "twinkle"]

LCDataType = t.Dict[InstrumentType, LightCurveInstrumentData]


class LightCurveData(Logger, Writeable):
    """Class holding data for a lightcurve from an instrument."""

    availableInstruments = ["wfc3", "spitzer", "stis", "twinkle"]  # noqa: N815

    @classmethod
    def fromInstrumentName(  # noqa: N802
        cls,
        name: InstrumentType,
        lc_data: LCDataType,
    ):
        log = logging.getLogger()
        if name.lower() in ("wfc3",):
            return cls(lc_data, "wfc3", (1.1, 1.8))
        elif name.lower() in ("spitzer",):
            return cls(lc_data, "spitzer", (3.4, 8.2))
        elif name.lower() in ("stis",):
            return cls(lc_data, "stis", (0.3, 1.0))
        elif name.lower() in ("twinkle",):
            return cls(lc_data, "twinkle", (0.4, 4.5))
        else:
            log.error(
                "LightCurve of instrument %s not recognized" " or implemented", name
            )
            raise KeyError

    def __init__(
        self,
        lc_data: LCDataType,
        instrument_name: InstrumentType,
        wavelength_region: t.Tuple[float, float],
    ):
        """Initialize.

        Parameters
        ----------
        lc_data:
            Lightcurve data
        instrument_name:
            Name of instrument
        wavelength_region:
            Wavelength region of instrument

        Raises
        ------
        KeyError
            If instrument name not found in lightcurve data


        """
        super().__init__(self.__class__.__name__)
        self._instrument_name = instrument_name
        # new version
        if self._instrument_name not in lc_data:
            self.error(
                "Instrument with key %s not found in pickled lightcurve" " file",
                self._instrument_name,
            )
            raise KeyError()

        self._wavelength_region = wavelength_region
        self._load_data(lc_data)

    def _load_data(self, lc_data: LCDataType):
        """Load data from lightcurve data."""
        # new
        self._time_series = lc_data[self._instrument_name]["time_series"]

        # new
        self._raw_data = lc_data[self._instrument_name]["data"][:, :, 0]
        self._data_std = lc_data[self._instrument_name]["data"][:, :, 1]

        self._max_nfactor = np.max(self._raw_data, axis=1)
        self._min_nfactor = np.min(self._raw_data, axis=1)

    @property
    def instrumentName(self) -> InstrumentType:  # noqa: N802
        """Instrument name."""
        return self._instrument_name

    @property
    def wavelengthRegion(self) -> t.Tuple[float, float]:  # noqa: N802
        """Wavelength region of instrument."""
        return self._wavelength_region

    @property
    def timeSeries(self) -> npt.NDArray[np.float64]:  # noqa: N802
        """Time series of data."""
        return self._time_series

    @property
    def rawData(self) -> npt.NDArray[np.float64]:  # noqa: N802
        """Raw data."""
        return self._raw_data

    @property
    def dataError(self) -> npt.NDArray[np.float64]:  # noqa: N802
        """Lightcurve uncertainty."""
        return self._data_std

    @property
    def minNFactors(self) -> npt.NDArray[np.float64]:  # noqa: N802
        """Minimum number of factors."""
        return self._min_nfactor

    @property
    def maxNFactors(self) -> npt.NDArray[np.float64]:  # noqa: N802
        """Maximum number of factors."""
        return self._max_nfactor

    def write(self, output: OutputGroup) -> OutputGroup:
        """Write to output group."""
        lc_grp = output.create_group(self.instrumentName)

        lc_grp.write_array("raw_data", self.rawData)
        lc_grp.write_array("data_error", self.dataError)
        lc_grp.write_array("min_n_factors", self.minNFactors)
        lc_grp.write_array("max_n_factors", self.maxNFactors)
        lc_grp.write_array("time_series", self.timeSeries)
        lc_grp.write_scalar("min_wavelength", self.wavelengthRegion[0])
        lc_grp.write_scalar("max_wavelength", self.wavelengthRegion[1])

        return lc_grp



"""Opacity integration of collision-induced absorption."""
import typing as t

import numpy as np
import numpy.typing as npt

from taurex.cache import CIACache
from taurex.model import OneDForwardModel
from taurex.output import OutputGroup

from .contribution import Contribution

contribute_cia: t.Callable[
    [
        int,
        int,
        int,
        npt.NDArray[np.float64],
        npt.NDArray[np.float64],
        npt.NDArray[np.float64],
        int,
        int,
        int,
        npt.NDArray[np.float64],
    ],
    npt.NDArray[np.float64],
] = None


def contribute_cia_numpy(
    startk: int,
    endk: int,
    density_offset: int,
    sigma: npt.NDArray[np.float64],
    density: npt.NDArray[np.float64],
    path: npt.NDArray[np.float64],
    nlayers: int,
    ngrid: int,
    layer: int,
    tau: npt.NDArray[np.float64],
):
    """Integrate the optical depth for a given layer."""
    _path = path[startk:endk, None]
    _density = density[startk + density_offset : endk + density_offset, None]
    _sigma = sigma[startk + layer : endk + layer, :]

    tau[layer, :] += np.sum(_sigma * _path * _density * _density, axis=0)


def contribute_cia_numba(
    startk: int,
    endk: int,
    density_offset: int,
    sigma: npt.NDArray[np.float64],
    density: npt.NDArray[np.float64],
    path: npt.NDArray[np.float64],
    nlayers: int,
    ngrid: int,
    layer: int,
    tau: npt.NDArray[np.float64],
) -> npt.NDArray[np.float64]:
    """Collisionally induced absorption integration function.

    This has the form:

    .. math::

        \\tau_{\\lambda}(z) = \\int_{z_{0}}^{z_{1}}
            \\sigma(z') \\rho(z')^{2} dz',

    where :math:`z` is the layer, :math:`z_0` and :math:`z_1` are ``startK``
    and ``endK`` respectively. :math:`\\sigma` is the weighted
    cross-section ``sigma``. :math:`rho` is the ``density`` and
    :math:`dz'` is the integration path length ``path``


    Parameters
    ----------
    startK: int
        starting layer in integration

    endK: int
        last layer in integration

    density_offset: int
        Which part of the density profile to start from

    sigma: :obj:`array`
        cross-section

    density: array_like
        density profile of atmosphere

    path: array_like
        path-length or altitude gradient

    nlayers: int
        Total number of layers (unused)

    ngrid: int
        total number of grid points

    layer: int
        Which layer we currently on

    Returns
    -------
    tau : array_like
        optical depth (well almost you still need to do ``exp(-tau)`` yourself)

    """
    for k in range(startk, endk):
        _path = path[k]
        _density = density[k + density_offset]
        # for mol in range(nmols):
        for wn in range(ngrid):
            tau[layer, wn] += sigma[k + layer, wn] * _path * _density * _density


try:
    import numba

    contribute_cia = numba.jit(contribute_cia_numba, nopython=True, nogil=True)

except ImportError:
    # Non numba version.
    contribute_cia = contribute_cia_numpy


class CIAContribution(Contribution):
    """Computes the CIA contribution to the optical depth.

    CIA is collisionally induced absorption.
    """

    def __init__(self, cia_pairs: t.Optional[t.List[str]] = None) -> None:
        """Initialize CIA.

        Parameters
        ----------
        cia_pairs: :obj:`list` of str
            list of molecule pairs of the form ``mol1-mol2``
            e.g. ``H2-He``

        """
        super().__init__("CIA")
        self._cia_pairs = cia_pairs

        self._cia_cache = CIACache()
        if self._cia_pairs is None:
            self._cia_pairs = []

    @property
    def ciaPairs(self) -> t.Sequence[str]:  # noqa: N802
        """Returns list of molecular pairs involved.

        Returns
        -------
        :obj:`list` of str
            list of molecule pairs of the form ``mol1-mol2``
            e.g. ``H2-He``
        """

        return self._cia_pairs

    @ciaPairs.setter
    def ciaPairs(self, value: t.List[str]) -> None:  # noqa: N806,N802
        """Sets list of molecular pairs involved.

        Parameters
        ----------
        value: :obj:`list` of str
            list of molecule pairs of the form ``mol1-mol2``
            e.g. ``H2-He``

        """
        self._cia_pairs = value

    def contribute(
        self,
        model: OneDForwardModel,
        start_layer: int,
        end_layer: int,
        density_offset: int,
        layer: int,
        density: npt.NDArray[np.float64],
        tau: npt.NDArray[np.float64],
        path_length: t.Optional[npt.NDArray[np.float64]] = None,
    ) -> None:
        """Integrate the optical depth for a given layer."""
        if self._total_cia > 0:
            contribute_cia(
                start_layer,
                end_layer,
                density_offset,
                self.sigma_xsec,
                density,
                path_length,
                self._nlayers,
                self._ngrid,
                layer,
                tau,
            )

    def prepare_each(
        self, model: OneDForwardModel, wngrid: npt.NDArray[np.float64]
    ) -> t.Generator[t.Tuple[str, npt.NDArray[np.float64]], None, None]:
        """Computes and weighs cross-section for a single pair of molecules

        Parameters
        ----------
        model: :class:`~taurex.model.model.ForwardModel`
            Forward model

        wngrid: :obj:`array`
            Wavenumber grid


        Yields
        ------
        component: :obj:`tuple` of type (str, :obj:`array`)
            Molecular pair and the weighted cia opacity.

        """
        self._total_cia = len(self.ciaPairs)
        self._nlayers = model.nLayers
        self._ngrid = wngrid.shape[0]
        self.info("Computing CIA ")

        sigma_cia = np.zeros(shape=(model.nLayers, wngrid.shape[0]))

        chemistry = model.chemistry

        for pair_name in self.ciaPairs:
            cia = self._cia_cache[pair_name]
            sigma_cia[...] = 0.0

            cia_factor = chemistry.get_gas_mix_profile(
                cia.pairOne
            ) * chemistry.get_gas_mix_profile(cia.pairTwo)

            for idx_layer, temperature in enumerate(model.temperatureProfile):
                _cia_xsec = cia.cia(temperature, wngrid)
                sigma_cia[idx_layer] += _cia_xsec * cia_factor[idx_layer]
            self.sigma_xsec = sigma_cia
            yield pair_name, sigma_cia

    def write(self, output: OutputGroup) -> OutputGroup:
        """Write output to file."""
        contrib = super().write(output)
        if len(self.ciaPairs) > 0:
            contrib.write_string_array("cia_pairs", self.ciaPairs)
        return contrib

    @classmethod
    def input_keywords(cls) -> t.Tuple[str]:
        """Return list of input keywords for CIA."""
        return ("CIA",)



"""Writeable class."""
from abc import ABCMeta, abstractmethod


class Writeable(metaclass=ABCMeta):
    """Abstract class for writeable objects."""

    @abstractmethod
    def write(self, output):
        pass



"""Module dealing with observed lightcurves."""

import typing as t

import numpy as np
import numpy.typing as npt

from taurex.binning.lightcurvebinner import LightcurveBinner
from taurex.model.lightcurve.lightcurvedata import LCDataType, LightCurveData
from taurex.output import OutputGroup
from taurex.types import PathLike

from .spectrum import BaseSpectrum


class ObservedLightCurve(BaseSpectrum):
    """Loads an observed lightcurve from a pickle file."""

    def __init__(self, filename: t.Optional[PathLike] = None):
        """Initialize lightcurve.

        Parameters
        ----------
        filename:
            Filename of lightcurve pickle data, by default None

        """
        super().__init__("observed_lightcurve")

        import pickle  # noqa: S403

        with open(filename, "rb") as f:
            lc_data = pickle.load(f, encoding="latin1")  # noqa: S301
        # new version
        self.obs_spectrum = np.empty(shape=(len(lc_data["obs_spectrum"][:, 0]), 4))

        # new version
        self.obs_spectrum[:, 0] = lc_data["obs_spectrum"][:, 0]
        self.obs_spectrum[:, 1] = lc_data["obs_spectrum"][:, 1]
        self.obs_spectrum[:, 2] = lc_data["obs_spectrum"][:, 2]
        self.obs_spectrum[:, 3] = lc_data["obs_spectrum"][:, 3]

        self._spec, self._std = self._load_data_file(lc_data)

    def create_binner(self) -> LightcurveBinner:
        """Creates the appropriate binning object."""
        return LightcurveBinner()

    def _load_data_file(
        self, lc_data: LCDataType
    ) -> t.Tuple[npt.NDArray[np.float64], npt.NDArray[np.float64]]:
        """Load and combine data from different instruments.

        Parameters
        ----------
        lc_data:
            Lightcurve data

        Returns
        -------
        combine_lc:
            Combined lightcurve.
        combine_lc_std:
            Combined lightcurve uncertainty.


        """
        raw_data = []
        data_std = []
        wngrid_min = []

        for ins in LightCurveData.availableInstruments:
            # new version
            # raw data includes data and datastd.
            if ins in lc_data:
                wngrid_min.append(lc_data[ins]["wl_grid"].min())
                raw_data.append(lc_data[ins]["data"][:, :, 0])
                data_std.append(lc_data[ins]["data"][:, :, 1])
        wngrid_min, raw_data, data_std = (
            list(t)
            for t in zip(
                *sorted(
                    zip(wngrid_min, raw_data, data_std),
                    key=lambda x: x[0],
                    reverse=True,
                )
            )
        )
        return np.concatenate(raw_data), np.concatenate(data_std)

    @property
    def spectrum(self) -> npt.NDArray[np.float64]:
        """Return a Light curve `spectrum`.

        Spectrum is not a true spectrum but in the context of Taurex it is
        seen as one to a retrieval.

        The lightcurve spectrum comes in the form of multiple lightcurves
        stuck together into
        one long spectrum. The number of lightcurves is equal to the number of
        bins in :func:`wavelengthGrid`.

        """
        return self._spec

    @property
    def rawData(self) -> npt.NDArray[np.float64]:  # noqa: N802
        """Raw lightcurve data read from file

        Returns
        -------
        lc_data : :obj:`array`

        """
        self.obs_spectrum

    @property
    def wavelengthGrid(self) -> npt.NDArray[np.float64]:  # noqa: N802
        """Returns wavelength grid in microns

        Returns
        -------
        wlgrid : :obj:`array`

        """
        return self.obs_spectrum[:, 0]

    @property
    def binEdges(self) -> npt.NDArray[np.float64]:  # noqa: N802
        """Returns bin edges for wavelength grid.

        Returns
        -------
        out : :obj:`array`

        """
        return self.obs_spectrum[:, 3]

    @property
    def binWidths(self) -> npt.NDArray[np.float64]:  # noqa: N802
        """Widths for each bin in wavelength grid.

        Returns
        -------
        out : :obj:`array`

        """
        return np.zeros(2)

    @property
    def errorBar(self) -> npt.NDArray[np.float64]:  # noqa: N802
        """Uncertainty of lightcurve spectrum.

        Returns
        -------
        err : :obj:`array`
            Error at each point in lightcurve spectrum

        """
        return self._std

    def write(self, output: OutputGroup) -> OutputGroup:
        """Write to output group."""
        output.write_array("wlgrid", self.wavelengthGrid)
        output.write_array("spectrum", self.obs_spectrum[:, 1])
        output.write_array("lightcurve", self.spectrum)
        output.write_array("binedges", self.binEdges)
        output.write_array("binwidths", self.binWidths)
        output.write_array("errorbars", self.obs_spectrum[:, 2])
        output.write_array("lightcurve_errorbars", self.errorBar)

        return output

    @classmethod
    def input_keywords(cls) -> t.Tuple[str, ...]:
        """Input keywords for this class."""
        return ("lightcurve",)



"""Instrument loaded from file."""
import math
import typing as t

import numpy as np

from taurex.binning import BinDownType, FluxBinner
from taurex.model import ForwardModel
from taurex.types import ModelOutputType, PathLike
from taurex.util import wnwidth_to_wlwidth

from .instrument import Instrument


class InstrumentFile(Instrument):
    """Loads a 2-3 column file

    The first column is the wavelength grid, the second column is the noise
    and the third column is the width of the wavelength bin. If the third column
    is not present, the width is computed from the wavelength grid.


    """

    def __init__(
        self,
        filename: t.Optional[PathLike] = None,
        delimiter: t.Optional[str] = None,
        skiprows: t.Optional[int] = 0,
        use_cols: t.Optional[t.Tuple[int, ...]] = None,
    ) -> None:
        super().__init__()

        self._spectrum = np.loadtxt(
            filename, skiprows=skiprows, delimiter=delimiter, usecols=use_cols
        )

        self._wlgrid = self._spectrum[:, 0]

        sortedwl = self._wlgrid.argsort()[::-1]

        self._wlgrid = self._wlgrid[sortedwl]

        self._wngrid = 10000 / self._wlgrid

        self._noise = self._spectrum[sortedwl, 1]

        try:
            self._wlwidths = self._spectrum[sortedwl, 2]
        except IndexError:
            from taurex.util import compute_bin_edges

            self._wlwidths - compute_bin_edges(self._wlgrid)[-1]

        self.create_wn_widths()

        self._binner = FluxBinner(self._wngrid, wngrid_width=self._wnwidths)

    def create_wn_widths(self) -> None:
        """Covert wavelength widths to wavenumber widths."""
        self._wnwidths = wnwidth_to_wlwidth(self._wlgrid, self._wlwidths)

    def model_noise(
        self,
        model: ForwardModel,
        model_res: t.Optional[ModelOutputType] = None,
        num_observations: t.Optional[int] = 1,
    ) -> BinDownType:
        """Attach noise to forward model.

        Parameters
        ----------

        model:
            Forward model to pass.

        model_res:
            Result from :func:`~taurex.model.model.ForwardModel.model`

        num_observations:
            Number of observations to simulate

        """
        if model_res is None:
            model_res = model.model()

        wngrid, spectrum, _, grid_width = self._binner.bin_model(model_res)

        return wngrid, spectrum, self._noise / math.sqrt(num_observations), grid_width

    @classmethod
    def input_keywords(cls) -> t.Tuple[str, ...]:
        return (
            "file",
            "fromfile",
        )



"""Handling of .pickle/.db CIA files."""
import pickle  # noqa: S403
import typing as t
from pathlib import Path

import numpy as np
import numpy.typing as npt

from taurex.util.math import interp_lin_only

from .cia import CIA


class PickleCIA(CIA):
    """
    Class for using pickled (``.db``) collisionally induced absorptions
    Very simple since the format is simple


    Parameters
    ----------
    filename : str
        Path to pickle

    pair_name : str , optional
        Whilst the name of the pair is determined by the pickle filename
        since these can be different you can optionally force the name through
        this parameter

    """

    def __init__(self, filename: str, pair_name: t.Optional[np.float64] = None):
        if pair_name is None:
            pair_name = Path(filename).stem

        super().__init__("PickleCIA", pair_name)

        self._filename = filename
        self._molecule_name = None
        self._spec_dict = None
        self._load_pickle_file(filename)

    def _load_pickle_file(self, filename: str) -> None:
        """Loads pickle file.

        Parameters
        ----------
        filename : str
            Path to pickle cia file

        """

        # Load the pickle file
        self.info("Loading cia cross section from %s", filename)
        with open(filename, "rb") as f:
            self._spec_dict = pickle.load(f, encoding="latin1")  # noqa: S301

        self._wavenumber_grid = self._spec_dict["wno"]
        self._temperature_grid = self._spec_dict["t"]
        self._xsec_grid = self._spec_dict["xsecarr"]

    @property
    def wavenumberGrid(self) -> npt.NDArray[np.float64]:  # noqa: N802
        """

        Returns
        -------
        :obj:`array`
            Native wavenumber grid

        """
        return self._wavenumber_grid

    @property
    def temperatureGrid(self) -> npt.NDArray[np.float64]:  # noqa: N802
        """

        Returns
        -------
        :obj:`array`
            Native temperature grid in Kelvin

        """

        return self._temperature_grid

    def find_closest_temperature_index(self, temperature: float) -> t.Tuple[int, int]:
        """
        Finds the nearest indices for a particular temperature

        Parameters
        ----------
        temperature : float
            Temeprature in Kelvin

        Returns
        -------
        t_min : int
            index on temprature grid to the left of ``temperature``

        t_max : int
            index on temprature grid to the right of ``temperature``

        """
        from taurex.util import find_closest_pair

        t_min, t_max = find_closest_pair(self.temperatureGrid, temperature)
        return t_min, t_max

    def interp_linear_grid(
        self, temperature: float, t_idx_min: int, t_idx_max: int
    ) -> float:
        """
        For a given temperature and indicies. Interpolate the cross-sections
        linearly from temperature grid to temperature ``T``

        Parameters
        ----------
        temperature : float
            Temeprature in Kelvin

        t_min : int
            index on temprature grid to the left of ``temperature``

        t_max : int
            index on temprature grid to the right of ``temperature``

        Returns
        -------
        out : :obj:`array`
            Interpolated cross-section

        """
        if temperature > self._temperature_grid.max():
            return self._xsec_grid[-1]
        elif temperature < self._temperature_grid.min():
            return self._xsec_grid[0]

        temp_max = self._temperature_grid[t_idx_max]
        temp_min = self._temperature_grid[t_idx_min]
        fx0 = self._xsec_grid[t_idx_min]
        fx1 = self._xsec_grid[t_idx_max]

        return interp_lin_only(fx0, fx1, temperature, temp_min, temp_max)

    def compute_cia(self, temperature: float) -> npt.NDArray[np.float64]:
        """
        Computes the collisionally induced absorption cross-section
        using our native temperature and cross-section grids

        Parameters
        ----------
        temperature : float
            Temperature in Kelvin

        Returns
        -------
        out : :obj:`array`
            Temperature interpolated cross-section

        """
        indicies = self.find_closest_temperature_index(temperature)
        return self.interp_linear_grid(temperature, *indicies)



"""Gas profile in Power law."""

import typing as t

import numpy as np
import numpy.typing as npt

from taurex.output import OutputGroup
from taurex.util import molecule_texlabel

from .gas import Gas

ProfileType = t.Literal["auto", "H2", "H2O", "TiO", "VO", "H-", "Na", "K"]
"""Profile type for power law."""

KNOWN_MOLECULES = ["H2", "H2O", "TiO", "VO", "H-", "Na", "K"]
"""Known molecules for power law."""


class PowerGas(Gas):
    """Gas profile in Power law.

    This is a profile adapted for HJ (T > 2000 K) which takes
    into account upper atm mix reduction. Laws taken from Parmentier (2018)

    """

    def __init__(
        self,
        molecule_name: t.Optional[str] = "H2O",
        profile_type: ProfileType = "auto",
        mix_ratio_surface: t.Optional[float] = None,
        alpha: t.Optional[float] = None,
        beta: t.Optional[float] = None,
        gamma: t.Optional[float] = None,
    ):
        """Initialize power gas profile.

        Parameters
        -----------
        molecule_name : str
            Name of molecule (for the xsec)

        profile_type : str , optional
            name of the molecule to take the profile from: 'H2','H2O','TiO',
                'VO', 'H-', 'Na', 'K'
            by default it uses 'auto' to get the same profile as for molecule_name

        mix_ratio_surface : float , optional
            Mixing ratio of the molecule on the planet surface

        alpha : float , optional
            pressure dependance coefficient approx 10^alpha

        beta : float , optional
            temperature dependance coefficient approx 10^(beta/T)

        gamma : float , optional
            scale coefficient

        Raises
        -------
        ValueError
            if the molecule is not known and profile_type is not ``auto``

        """
        super().__init__("PowerGas", molecule_name=molecule_name)
        if profile_type == "auto":
            self._profile_type = molecule_name
        elif profile_type not in KNOWN_MOLECULES:
            raise ValueError(
                f"Unknown profile type {profile_type}, "
                f"please choose from {KNOWN_MOLECULES} if not auto"
            )
        else:
            self._profile_type = profile_type

        self._mix_surface = mix_ratio_surface
        self._alpha = alpha
        self._beta = beta
        self._gamma = gamma
        self._mix_profile = None
        self.add_surface_param()
        self.add_alpha_param()
        self.add_beta_param()
        self.add_gamma_param()

    @property
    def mixProfile(self) -> npt.NDArray[np.float64]:  # noqa: N802
        """Mixing ratio profile."""
        return self._mix_profile

    @property
    def mixRatioSurface(self) -> float:  # noqa: N802
        """Abundance on the planets surface."""
        return self._mix_surface

    @property
    def alpha(self) -> float:
        """Pressure dependance coefficient."""
        return self._alpha

    @property
    def beta(self) -> float:
        """Temperature dependance coefficient."""
        return self._beta

    @property
    def gamma(self) -> float:
        """Scale coefficient."""
        return self._gamma

    @mixRatioSurface.setter
    def mixRatioSurface(self, value: float):  # noqa: N802
        """Set the abundance on the planets surface."""
        self._mix_surface = value

    @alpha.setter
    def alpha(self, value: float):
        """Set the pressure dependance coefficient."""
        self._alpha = value

    @beta.setter
    def beta(self, value: float):
        """Set the temperature dependance coefficient."""
        self._beta = value

    @gamma.setter
    def gamma(self, value: float):
        """Set the scale coefficient."""
        self._gamma = value

    def add_surface_param(self):
        """Add surface parameter to fit."""
        param_name = self.molecule
        param_tex = molecule_texlabel(param_name)

        param_surface = f"{param_name}_surface"
        param_surf_tex = f"{param_tex}_surface"

        def read_surf(self):
            return self._mix_surface

        def write_surf(self, value):
            self._mix_surface = value

        fget_surf = read_surf
        fset_surf = write_surf

        bounds = [1.0e-12, 0.1]

        default_fit = False
        self.add_fittable_param(
            param_surface,
            param_surf_tex,
            fget_surf,
            fset_surf,
            "log",
            default_fit,
            bounds,
        )

    def add_alpha_param(self):
        """Add alpha parameter to fit."""
        param_name = self.molecule
        param_tex = molecule_texlabel(param_name)

        param_alpha = f"{param_name}_alpha"
        param_alpha_tex = f"{param_tex}_alpha"

        def read_alpha(self):
            return self._alpha

        def write_alpha(self, value):
            self._alpha = value

        fget_alpha = read_alpha
        fset_alpha = write_alpha

        bounds = [0.5, 2.5]

        default_fit = False
        self.add_fittable_param(
            param_alpha,
            param_alpha_tex,
            fget_alpha,
            fset_alpha,
            "linear",
            default_fit,
            bounds,
        )

    def add_beta_param(self):
        """Add beta parameter to fit."""
        param_name = self.molecule
        param_tex = molecule_texlabel(param_name)

        param_beta = f"{param_name}_beta"
        param_beta_tex = f"{param_tex}_beta"

        def read_beta(self):
            return self._beta

        def write_beta(self, value):
            self._beta = value

        fget_beta = read_beta
        fset_beta = write_beta

        bounds = [1e4, 6e4]

        default_fit = False
        self.add_fittable_param(
            param_beta, param_beta_tex, fget_beta, fset_beta, "log", default_fit, bounds
        )

    def add_gamma_param(self):
        """Add gamma parameter to fit."""
        param_name = self.molecule
        param_tex = molecule_texlabel(param_name)

        param_gamma = f"{param_name}_gamma"
        param_gamma_tex = f"{param_tex}_gamma"

        def read_gamma(self):
            return self._gamma

        def write_gamma(self, value):
            self._gamma = value

        fget_gamma = read_gamma
        fset_gamma = write_gamma

        bounds = [5, 25]

        default_fit = False
        self.add_fittable_param(
            param_gamma,
            param_gamma_tex,
            fget_gamma,
            fset_gamma,
            "linear",
            default_fit,
            bounds,
        )

    def check_known(
        self,
        molecule_name: t.Literal["H2", "H2O", "TiO", "VO", "H-", "Na", "K"],
    ) -> t.Tuple[float, float, float, float]:
        """Check if the molecule is known and return the coefficients.

        Parameters
        -----------
        molecule_name : str
            name of the molecule to take the profile from: 'H2','H2O','TiO',
                'VO', 'H-', 'Na', 'K'

        Returns
        --------
        a_coeff : float
            power coefficient for pressure or None
        b_coeff : float
            power coefficient for temperature or None
        g_coeff : float
            scale coefficient or None
        big_a_coeff : float
            mixing ratio at the surface or None

        Raises
        -------
        ValueError
            if the molecule is not known

        """
        a_coeff = np.array([1.0, 2.0, 1.6, 1.5, 0.6, 0.6, 0.6])
        b_coeff = np.array([2.41, 4.83, 5.94, 5.4, -0.14, 1.89, 1.28]) * 1e-4
        g_coeff = np.array([6.5, 15.9, 23.0, 23.8, 7.7, 12.2, 12.7])
        big_a_coeff = np.power(10, np.array([-0.1, -3.3, -7.1, -9.2, -8.3, -5.5, -7.1]))
        if molecule_name in KNOWN_MOLECULES:
            i = KNOWN_MOLECULES.index(molecule_name)
            self.debug(
                "%s %s %s %s %s", i, a_coeff[i], b_coeff[i], g_coeff[i], big_a_coeff[i]
            )
            return a_coeff[i], b_coeff[i], g_coeff[i], big_a_coeff[i]
        else:
            raise ValueError(
                f"molecule {molecule_name} is not known, "
                f"please choose from {KNOWN_MOLECULES}"
            )

    def initialize_profile(
        self,
        nlayers: int,
        temperature_profile: npt.NDArray[np.float64],
        pressure_profile: npt.NDArray[np.float64],
        altitude_profile: t.Optional[npt.NDArray[np.float64]] = None,
    ) -> None:
        """Initialize the profile.

        Parameters
        -----------
        nlayers : int
            number of layers
        temperature_profile : np.ndarray
            temperature profile
        pressure_profile : np.ndarray
            pressure profile
        altitude_profile : np.ndarray , optional
            altitude profile, deprecated

        """
        self._mix_profile = np.zeros(nlayers)
        molecule_name = self._profile_type
        coeffs = None, None, None, None
        if molecule_name in KNOWN_MOLECULES:
            coeffs = self.check_known(molecule_name=molecule_name)

        mix_surface = self._mix_surface
        alpha = self._alpha
        beta = self._beta
        gamma = self._gamma
        ad = []

        if self._mix_surface is None:
            if coeffs[3] is not None:
                mix_surface = coeffs[3]
            else:
                self.error("molecule %s has a missing power coefficient", molecule_name)
                raise ValueError
        if self._alpha is None:
            if coeffs[0] is not None:
                alpha = coeffs[0]
            else:
                self.error("molecule %s has a missing power coefficient", molecule_name)
                raise ValueError
        if self._beta is None:
            if coeffs[1] is not None:
                beta = coeffs[1]
            else:
                self.error("molecule %s has a missing power coefficient", molecule_name)
                raise ValueError
        if self._gamma is None:
            if coeffs[2] is not None:
                gamma = coeffs[2]
            else:
                self.error("molecule %s has a missing power coefficient", molecule_name)
                raise ValueError
        # print('coeffs: ', coeffs)
        # print('a, b, g, minx_S: ', alpha, beta, gamma, mix_surface)

        # Ad = alpha * np.log10(pressure_profile) + beta / temperature_profile + gamma

        pressure = pressure_profile * 1e-5  # convert pressure to bar
        ad = (
            np.power(10, gamma * (-1))
            * np.power(pressure, alpha)
            * np.power(10, beta / temperature_profile)
        )
        mix = 1 / np.sqrt(mix_surface) + 1 / np.sqrt(ad)
        mix = np.power(1 / mix, 2)

        self._mix_profile = mix

    def write(self, output: OutputGroup) -> OutputGroup:
        """Write the gas profile to output."""
        gas_entry = super().write(output)
        gas_entry.write_scalar("alpha", self.alpha)
        gas_entry.write_scalar("mix_ratio_surface", self.mixRatioSurface)
        gas_entry.write_scalar("beta", self.beta)
        gas_entry.write_scalar("gamma", self.gamma)

        return gas_entry

    @classmethod
    def input_keywords(cls) -> t.Tuple[str, str, str]:
        """Return input keywords for power law gas profile."""
        return (
            "power",
            "powerchemistry",
            "parmentier",
        )



"""Outputs using HDF5 format."""

import datetime
import typing as t

import h5py
import numpy as np

from taurex.mpi import only_master_rank
from taurex.types import ArrayType, ScalarType

from .output import MetadataType, Output, OutputGroup


class HDF5OutputGroup(OutputGroup):
    """Stores output data in the HDF5 hierarchical structure."""

    def __init__(
        self, entry: t.Optional[h5py.Group] = None, name: t.Optional[str] = None
    ) -> None:
        """Initialize HDF5 output group."""
        super().__init__(name=name or "HDF5Group")
        self._entry = entry

    @only_master_rank
    def write_array(
        self,
        array_name: str,
        array: ArrayType,
        metadata: t.Optional[MetadataType] = None,
    ) -> None:
        array = np.array(array)
        ds = self._entry.create_dataset(
            str(array_name), data=array, shape=array.shape, dtype=array.dtype
        )
        if metadata:
            for k, v in metadata.items():
                ds.attrs[k] = v

    @only_master_rank
    def write_scalar(
        self,
        scalar_name: str,
        scalar: ScalarType,
        metadata: t.Optional[MetadataType] = None,
    ) -> None:
        ds = self._entry.create_dataset(str(scalar_name), data=scalar)
        if metadata:
            for k, v in metadata.items():
                ds.attrs[k] = v

    @only_master_rank
    def write_string(
        self, string_name: str, string: str, metadata: t.Optional[MetadataType] = None
    ) -> None:
        ds = self._entry.create_dataset(str(string_name), data=string)
        if metadata:
            for k, v in metadata.items():
                ds.attrs[k] = v

    def create_group(self, group_name: str) -> "HDF5OutputGroup":
        """Create a group."""
        entry = None
        if self._entry is not None:
            entry = self._entry.create_group(str(group_name))
            return HDF5OutputGroup(entry=entry, name=group_name)
        else:
            return HDF5OutputGroup(name=group_name, entry=None)

    @only_master_rank
    def write_string_array(
        self,
        string_name: str,
        string_array: t.Sequence[str],
        metadata: t.Dict[str, t.Any] = None,
    ) -> None:
        """Write a string array to the output."""
        ascii_lst = [n.encode("ascii", "ignore") for n in string_array]
        ds = self._entry.create_dataset(
            str(string_name), (len(ascii_lst), 1), "S64", ascii_lst
        )

        if metadata:
            for k, v in metadata.items():
                ds.attrs[k] = v


class HDF5Output(Output, HDF5OutputGroup):
    """Output using HDF5 format."""

    def __init__(self, filename, append=False):
        Output.__init__(self, "HDF5Output")
        HDF5OutputGroup.__init__(self, name="HDF5Output", entry=None)

        self.filename = filename
        self._append = append
        self.fd: h5py.File = None

    @only_master_rank
    def open(self) -> None:
        """Open file."""
        self.fd = self._open_file(self.filename)
        self._entry = self.fd

    def _open_file(self, fname: str) -> h5py.File:
        """Open file and write header."""
        from taurex._version import __version__

        mode = "w"
        if self._append:
            mode = "a"

        fd = h5py.File(fname, mode=mode)
        fd.attrs["file_name"] = fname
        fd.attrs["file_time"] = datetime.datetime.now().isoformat()
        fd.attrs["creator"] = self.__class__.__name__
        fd.attrs["HDF5_Version"] = h5py.version.hdf5_version
        fd.attrs["h5py_version"] = h5py.version.version
        fd.attrs["program_name"] = "TauREx"
        fd.attrs["program_version"] = __version__
        return fd

    def close(self) -> None:
        """Close file."""
        if self.fd:
            self.fd.flush()
            self.fd.close()
        self.fd = None



"""Opacities from Exo_Transmit.

Exo_Transmit is a code for calculating transmission spectra for exoplanet
atmospheres of varied composition. Its opacities are based on HITRAN and
have the format `opac{molecule}.dat`.
"""
import pathlib
import typing as t

import numpy as np
import numpy.typing as npt

from taurex.opacity.interpolateopacity import InterpModeType, InterpolatingOpacity
from taurex.types import PathLike


class ExoTransmitOpacity(InterpolatingOpacity):
    """Opacity from Exo_Transmit."""

    @classmethod
    def discover(
        cls,
    ) -> t.List[t.Tuple[str, t.Tuple[pathlib.Path, InterpModeType]]]:
        """Discover opacities from Exo_Transmit.

        They have the format `opac{molecule}.dat`.
        So we try to search for them in the xsec_path.


        """
        import pathlib

        from taurex.cache import GlobalCache
        from taurex.util import sanitize_molecule_string

        path = GlobalCache()["xsec_path"]
        if path is None:
            return []
        path = pathlib.Path(path)

        files = path.glob("opac*.dat")

        discovery = []

        interp = GlobalCache()["xsec_interpolation"] or "linear"

        for f in files:
            mol_name = sanitize_molecule_string(f.stem[4:])
            discovery.append((mol_name, [f, interp]))

        return discovery

    def __init__(
        self,
        filename: PathLike,
        interpolation_mode: t.Optional[InterpModeType] = "linear",
    ):
        """Initialize opacity from Exo_Transmit.

        Parameters
        ----------
        filename: PathLike
            Path to opacity file.

        interpolation_mode: str, optional
            Interpolation mode. Either 'linear' or 'exp'.


        Raises
        ------
        FileNotFoundError
            If file is not found.

        """
        super().__init__(
            f"ExoOpacity:{pathlib.Path(filename).stem[4:]}",
            interpolation_mode=interpolation_mode,
        )

        self._filename = pathlib.Path(filename)
        if not self._filename.exists():
            raise FileNotFoundError(f"Could not find {self._filename}")
        self._molecule_name = pathlib.Path(filename).stem[4:]
        self._load_exo_transmit(filename)

    def _load_exo_transmit(self, filename: pathlib.Path):
        """Load opacity from Exo_Transmit file.

        Parameters
        ----------
        filename: PathLike
            Path to opacity file.

        """
        self.debug(f"Loading opacity from {filename}")

        with open(filename) as f:
            lines = f.readlines()

        self._temperature_grid = np.array(
            [float(col) for col in lines[0].split()]
        )  # *t_conversion

        self._pressure_grid = np.array([float(col) for col in lines[1].split()]) * 1e5

        self._min_pressure = self._pressure_grid.min()
        self._max_pressure = self._pressure_grid.max()
        self._min_temperature = self._temperature_grid.min()
        self._max_temperature = self._temperature_grid.max()

        wn_grid = []

        for ln in lines[2:]:
            arr = np.array([float(col) for col in ln.split()])
            if arr.shape[0] == 1:
                wn_grid.append(10000 * 1e-6 / arr[0])

        wn_grid = np.array(wn_grid)

        grid_sort = wn_grid.argsort()
        self._wavenumber_grid = wn_grid[grid_sort]

        pressure_count = 0
        lambda_count = -1
        self._xsec_grid = np.empty(
            shape=(
                self.pressureGrid.shape[0],
                self.temperatureGrid.shape[0],
                self.wavenumberGrid.shape[0],
            )
        )

        for ln in lines[2:]:
            arr = np.array([float(col) for col in ln.split()])
            if arr.shape[0] == 1:
                lambda_count += 1
                pressure_count = 0
            else:
                self._xsec_grid[pressure_count, :, lambda_count] = arr[1:] + 1e-60
                pressure_count += 1

        self._xsec_grid = self._xsec_grid[:, :, grid_sort] * 10000

    @property
    def wavenumberGrid(self) -> npt.NDArray[np.float64]:  # noqa: N802
        """Wavenumber grid."""
        return self._wavenumber_grid

    @property
    def temperatureGrid(self) -> npt.NDArray[np.float64]:  # noqa: N802
        """Temperature grid."""
        return self._temperature_grid

    @property
    def pressureGrid(self) -> npt.NDArray[np.float64]:  # noqa: N802
        """Pressure grid."""
        return self._pressure_grid

    @property
    def resolution(self) -> float:
        """Resolution of opacity."""
        return self._resolution

    @property
    def moleculeName(self) -> str:  # noqa: N802
        """Name of molecule."""
        return self._molecule_name

    @property
    def xsecGrid(self) -> npt.NDArray[np.float64]:  # noqa: N802
        """Opacity grid."""
        return self._xsec_grid

    BIBTEX_ENTRIES = [
        """
        @ARTICLE{2017PASP..129d4402K,
            author = {{Kempton}, Eliza M. -R. and {Lupu},
            Roxana and {Owusu-Asare}, Albert and {Slough}, Patrick and {Cale}, Bryson},
                title = "{Exo-Transmit: An Open-Source Code
                for Calculating Transmission Spectra for
                Exoplanet Atmospheres of Varied Composition}",
            journal = {Publications of the Astronomical Society of the Pacific},
            keywords = {Astrophysics - Earth and Planetary Astrophysics},
                year = 2017,
                month = apr,
            volume = {129},
            number = {974},
                pages = {044402},
                doi = {10.1088/1538-3873/aa61ef},
        archivePrefix = {arXiv},
            eprint = {1611.03871},
        primaryClass = {astro-ph.EP},
            adsurl = {https://ui.adsabs.harvard.edu/abs/2017PASP..129d4402K},
            adsnote = {Provided by the SAO/NASA Astrophysics Data System}
        }

        """,
    ]



"""Flat Cloud scattering."""
import typing as t

import numpy as np
import numpy.typing as npt

from taurex.data.fittable import fitparam
from taurex.model import OneDForwardModel
from taurex.output import OutputGroup

from .contribution import Contribution


class SimpleCloudsContribution(Contribution):
    """
    Optically thick cloud deck up to a certain height

    These have the form:

    .. math::
            \\tau(\\lambda,z) =
                \\begin{cases}
                \\infty       & \\quad \\text{if } P(z) >= P_{0}\\\\
                0            & \\quad \\text{if } P(z) < P_{0}
                \\end{cases}

    Where :math:`P_{0}` is the pressure at the top of the cloud-deck.


    """

    def __init__(self, clouds_pressure: t.Optional[float] = 1e3) -> None:
        """Initialize the cloud model.

        Parameters
        ----------
        clouds_pressure : float, optional
            Pressure at top of cloud deck, by default 1e3


        """
        super().__init__("SimpleClouds")
        self._cloud_pressure = clouds_pressure

    @property
    def order(self) -> int:
        return 3

    def contribute(
        self,
        model: OneDForwardModel,
        start_layer: int,
        end_layer: int,
        density_offset: int,
        layer: int,
        density: npt.NDArray[np.float64],
        tau: npt.NDArray[np.float64],
        path_length: t.Optional[npt.NDArray[np.float64]] = None,
    ):
        """Contribute the cloud opacity to the optical depth."""
        tau[layer] += self.sigma_xsec[layer, :]

    def prepare_each(
        self, model: OneDForwardModel, wngrid: npt.NDArray[np.float64]
    ) -> t.Generator[t.Tuple[str, npt.NDArray[np.float64]], None, None]:
        """Compute cross-section that is infinitely absorbing.


        Absorption occurs up to a certain height.

        Parameters
        ----------
        model: :class:`~taurex.model.model.ForwardModel`
            Forward model

        wngrid: :obj:`array`
            Wavenumber grid

        Yields
        ------
        component: :obj:`tuple` of type (str, :obj:`array`)
            ``Clouds`` and opacity array.


        """

        contrib = np.zeros(
            shape=(
                model.nLayers,
                wngrid.shape[0],
            )
        )
        cloud_filtr = model.pressureProfile >= self._cloud_pressure
        contrib[cloud_filtr, :] = np.inf
        self._contrib = contrib
        yield "Clouds", self._contrib

    @fitparam(
        param_name="clouds_pressure",
        param_latex=r"$P_\mathrm{clouds}$",
        default_mode="log",
        default_fit=False,
        default_bounds=[1e-3, 1e6],
    )
    def cloudsPressure(self):  # noqa: N802
        """Cloud top pressure in Pascal."""
        return self._cloud_pressure

    @cloudsPressure.setter
    def cloudsPressure(self, value):  # noqa: N802
        """Cloud top pressure in Pascal."""
        self._cloud_pressure = value

    def write(self, output: OutputGroup):
        """Write the cloud pressure to the output."""
        contrib = super().write(output)
        contrib.write_scalar("clouds_pressure", self._cloud_pressure)
        return contrib

    @classmethod
    def input_keywords(cls) -> t.Tuple[str, str]:
        return (
            "SimpleClouds",
            "ThickClouds",
        )



"""Optimizer using multinest library."""

import os
import pathlib
import typing as t

import numpy as np
import numpy.typing as npt

from taurex import OutputSize
from taurex.model import ForwardModel
from taurex.mpi import barrier, get_rank
from taurex.output import OutputGroup
from taurex.spectrum import BaseSpectrum
from taurex.types import AnyValType, PathLike
from taurex.util import (
    quantile_corner,
    read_error_into_dict,
    read_table,
    recursively_save_dict_contents_to_output,
)

from .optimizer import FitParam, Optimizer

NestMarginalOutput = t.TypedDict(
    "NestMarginalOutput",
    {
        "median": float,
        "sigma": float,
        "1sigma": float,
        "2sigma": float,
        "3sigma": float,
        "5sigma": float,
        "q75%": float,
        "q25%": float,
        "q95%": float,
        "q01%": float,
        "q90%": float,
        "q10%": float,
    },
)
"""Marginal output."""
NestModeStatsOutput = t.TypedDict(
    "NestModeStatsOutput",
    {
        "Strictly Local Log-Evidence": t.Optional[float],
        "Strictly Local Log-Evidence error": t.Optional[float],
        "Local Log-Evidence": t.Optional[float],
        "Local Log-Evidence error": t.Optional[float],
        "mean": t.List[float],
        "sigma": t.List[float],
        "maximum": t.List[float],
        "maximum a posterior": t.List[float],
    },
)
"""Nested Mode stats output"""


NestStatsOutput = t.TypedDict(
    "NestStatsOutput",
    {
        "modes": t.List[NestModeStatsOutput],
        "marginals": t.List[NestMarginalOutput],
        "global evidence": t.Optional[float],
        "global evidence error": t.Optional[float],
    },
)


class NestFitParam(FitParam):
    """Fit parameter for multinest."""

    nest_map: float
    nest_median: float


class NestSolutionOutput(t.TypedDict):
    type: str
    local_logE: t.Tuple[float, float]  # noqa: N815
    weights: npt.NDArray[np.float64]
    tracedata: npt.NDArray[np.float64]
    fit_params: t.Dict[str, NestFitParam]


class NestOutputType(t.TypedDict):
    """Multinest output type"""

    NEST_stats: NestStatsOutput
    global_logE: t.Tuple[float, float]  # noqa: N815
    solutions: t.Dict[str, NestSolutionOutput]


class MultiNestOptimizer(Optimizer):
    def __init__(
        self,
        multi_nest_path: t.Optional[PathLike] = None,
        observed: t.Optional[BaseSpectrum] = None,
        model: t.Optional[ForwardModel] = None,
        sampling_efficiency: t.Optional[t.Literal["parameter"]] = "parameter",
        num_live_points: t.Optional[int] = 1500,
        max_iterations: t.Optional[int] = 0,
        search_multi_modes: t.Optional[bool] = True,
        num_params_cluster: t.Optional[int] = None,
        maximum_modes: t.Optional[int] = 100,
        constant_efficiency_mode: t.Optional[bool] = False,
        evidence_tolerance: t.Optional[float] = 0.5,
        mode_tolerance: t.Optional[float] = -1e90,
        importance_sampling: t.Optional[bool] = False,
        resume: t.Optional[bool] = False,
        only_finish: t.Optional[bool] = False,
        n_iter_before_update: t.Optional[int] = 100,
        multinest_prefix: t.Optional[str] = "1-",
        verbose_output: t.Optional[bool] = True,
        sigma_fraction: t.Optional[float] = 0.1,
    ):
        r"""Setup multinest optimizer.

        Parameters
        ----------
        multi_nest_path:
            Path to multinest output directory, by default None
        observed:
            Observed spectrum to optimize to
        model:
            Forward model to optimize
        sampling_efficiency:
            Sampling efficiency (parameter, ...)
        num_live_points:
            Number of live points
        max_iterations:
            Maximum no. of iterations (0=inf)
        search_multi_modes:
            Search for multiple modes
        num_params_cluster:
            Parameters on which to cluster
            e.g. if nclust_par = 3, it will cluster on the first 3 parameters only.
            if ncluster_par = -1 it clusters on all parameters
        maximum_modes:
            Maximum number of modes
        constant_efficiency_mode:
            Run in constant efficiency mode
        evidence_tolerance:
            Set log likelihood tolerance. :math:`\Delta \log Z <` evidence_tolerance
        mode_tolerance:
            Mode tolerance
        importance_sampling:
            Use importance nested sampling
        resume:
            Resume from previous run
        n_iter_before_update:
            Number of iterations before updating
        multinest_prefix:
            Prefix for multinest output files
        verbose_output:
            Verbose output
        sigma_fraction:
            Fraction of sigma to use for prior. Default is 10%
        """
        super().__init__("Multinest", observed, model, sigma_fraction)

        multi_nest_path = pathlib.Path(multi_nest_path)
        # sampling chains directory
        self.nest_path = "chains/"
        self.nclust_par = -1
        # sampling efficiency (parameter, ...)
        self.sampling_eff = sampling_efficiency
        # number of live points
        self.n_live_points = int(num_live_points)
        # maximum no. of iterations (0=inf)
        self.max_iter = int(max_iterations)
        # search for multiple modes
        self.multimodes = search_multi_modes
        self.n_iter_before_update = int(n_iter_before_update)
        # parameters on which to cluster, e.g. if nclust_par = 3, it will
        # cluster on the first 3 parameters only.
        # if ncluster_par = -1 it clusters on all parameters
        self.nclust_par = num_params_cluster
        # maximum number of modes
        self.max_modes = int(maximum_modes)
        # run in constant efficiency mode
        self.const_eff = constant_efficiency_mode
        self._only_finish = only_finish
        # set log likelihood tolerance. If change is smaller,
        # multinest will have converged
        self.evidence_tolerance = evidence_tolerance
        self.mode_tolerance = mode_tolerance
        # importance nested sampling
        self.imp_sampling = importance_sampling
        if self.imp_sampling:
            self.multimodes = False

        self.multinest_prefix = multinest_prefix

        if get_rank() == 0:
            if not multi_nest_path.exists():
                self.info("Directory %s does not exist, creating", multi_nest_path)
                multi_nest_path.mkdir(parents=True)
        barrier()

        # Convert to standard string path
        self.dir_multinest = multi_nest_path
        self.info("Found directory %s", self.dir_multinest)

        self.resume = resume
        self.verbose = verbose_output

    def compute_fit(self):
        """Compute the fit."""
        import json

        import pymultinest

        def multinest_loglike(cube, ndim, nparams):
            data = np.array([cube[i] for i in range(len(self.fitting_parameters))])
            return self.log_likelihood(data)

        def multinest_uniform_prior(cube, ndim, nparams):
            # prior distributions called by multinest. Implements a uniform prior
            # converting parameters from normalised grid to uniform prior
            # print(type(cube))
            data = np.array([cube[i] for i in range(nparams)])
            prior = self.prior_transform(data)
            for idx, c in enumerate(prior):
                cube[idx] = c

        ndim = len(self.fitting_parameters)
        self.warning(f"Number of dimensions {ndim}")
        self.warning(f"Fitting parameters {self.fitting_parameters}")

        ncluster = self.nclust_par

        if isinstance(ncluster, float):
            ncluster = int(ncluster)

        if ncluster is not None and ncluster <= 0:
            ncluster = None

        if ncluster is None:
            self.nclust_par = ndim  # For writing to output later on

        # Will help for live plotting
        with open(self.dir_multinest / f"{self.multinest_prefix}params.json", "w") as f:
            json.dump(self.fit_latex, f)

        # write param json file

        self.info("Beginning fit......")

        if not self._only_finish:
            pymultinest.run(
                LogLikelihood=multinest_loglike,
                Prior=multinest_uniform_prior,
                n_dims=ndim,
                multimodal=self.multimodes,
                n_clustering_params=ncluster,
                max_modes=self.max_modes,
                n_iter_before_update=self.n_iter_before_update,
                outputfiles_basename=str(self.dir_multinest / self.multinest_prefix),
                const_efficiency_mode=self.const_eff,
                importance_nested_sampling=self.imp_sampling,
                resume=self.resume,
                verbose=self.verbose,
                sampling_efficiency=self.sampling_eff,
                evidence_tolerance=self.evidence_tolerance,
                mode_tolerance=self.mode_tolerance,
                n_live_points=self.n_live_points,
                max_iter=self.max_iter,
            )

        self.info("Fit complete.....")

        self._multinest_output = self.store_nest_solutions()

        self.debug("Multinest output %s", self._multinest_output)

    def write_optimizer(self, output: OutputGroup) -> OutputGroup:
        """Write optimizer to output."""
        opt = super().write_optimizer(output)

        # sampling efficiency (parameter, ...)
        opt.write_scalar("sampling_eff ", self.sampling_eff)
        # number of live points
        opt.write_scalar("num_live_points", self.n_live_points)
        # maximum no. of iterations (0=inf)
        opt.write_scalar("max_iterations", self.max_iter)
        # search for multiple modes
        opt.write_scalar("search_multimodes", self.multimodes)
        # parameters on which to cluster, e.g. if nclust_par = 3,
        # it will cluster on the first 3 parameters only.
        # if ncluster_par = -1 it clusters on all parameters
        opt.write_scalar("nclust_parameter", self.nclust_par)
        # maximum number of modes
        opt.write_scalar("max_modes", self.max_modes)
        # run in constant efficiency mode
        opt.write_scalar("constant_efficiency", self.const_eff)
        # set log likelihood tolerance. If change is smaller,
        # multinest will have converged
        opt.write_scalar("evidence_tolerance", self.evidence_tolerance)
        opt.write_scalar("mode_tolerance", self.mode_tolerance)
        # importance nested sampling
        opt.write_scalar("importance_sampling ", self.imp_sampling)

        return opt

    def write_fit(self, output):
        fit = super().write_fit(output)

        if self._multinest_output:
            recursively_save_dict_contents_to_output(output, self._multinest_output)

        return fit

    # Laziness brought us to this point
    # This function is so big and I cannot be arsed to rewrite
    # this in a nicer way, if some angel does it
    # for me then I will buy them TWO beers.
    # Also I think pymultinest does most of this so why??!
    def store_nest_solutions(self) -> NestOutputType:  # noqa: C901
        """Store the multinest results."""
        import pymultinest

        self.warning("Store the multinest results")
        nest_out = {"solutions": {}}
        data = np.loadtxt(
            self.dir_multinest
            / f"{self.multinest_prefix}.txt"
            # os.path.join(self.dir_multinest, "{}.txt".format(self.multinest_prefix))
        )

        nest_analyser = pymultinest.Analyzer(
            n_params=len(self.fitting_parameters),
            outputfiles_basename=os.path.join(
                self.dir_multinest, self.multinest_prefix
            ),
        )
        nest_stats = nest_analyser.get_stats()
        nest_stats = t.cast(NestStatsOutput, nest_stats)
        nest_out["NEST_stats"] = nest_stats

        if "global evidence" in nest_stats:
            nest_out["global_logE"] = (
                nest_out["NEST_stats"]["global evidence"],
                nest_out["NEST_stats"]["global evidence error"],
            )

        # Bypass if run in multimodes = False. Pymultinest.Analyser does
        # not report means and sigmas in this case
        if len(nest_out["NEST_stats"]["modes"]) == 0:
            with open(
                self.dir_multinest
                / f"{self.multinest_prefix}stats.dat"
                # os.path.join(
                #     self.dir_multinest, "{}stats.dat".format(self.multinest_prefix)
                # )
            ) as file:
                lines = file.readlines()
            stats = {"modes": []}
            read_error_into_dict(lines[0], stats)

            text = "".join(lines[2:])
            # without INS:
            parts = text.split("\n\n")
            mode = {"index": 0}

            modelines = parts[0]
            tab = read_table(modelines, title="Parameter")
            mode["mean"] = tab[:, 1].tolist()
            mode["sigma"] = tab[:, 2].tolist()

            mode["maximum"] = read_table(parts[1], title="Parameter", d=None)[
                :, 1
            ].tolist()
            mode["maximum a posterior"] = read_table(
                parts[2], title="Parameter", d=None
            )[:, 1].tolist()

            if "Nested Sampling Global Log-Evidence".lower() in stats:
                mode["Local Log-Evidence".lower()] = stats[
                    "Nested Sampling Global Log-Evidence".lower()
                ]
                mode["Local Log-Evidence error".lower()] = stats[
                    "Nested Sampling Global Log-Evidence error".lower()
                ]
            else:
                mode["Local Log-Evidence".lower()] = stats[
                    "Nested Importance Sampling Global Log-Evidence".lower()
                ]
                mode["Local Log-Evidence error".lower()] = stats[
                    "Nested Importance Sampling Global Log-Evidence error".lower()
                ]

            mode["Strictly Local Log-Evidence".lower()] = mode[
                "Local Log-Evidence".lower()
            ]
            mode["Strictly Local Log-Evidence error".lower()] = mode[
                "Local Log-Evidence error".lower()
            ]

            nest_out["NEST_stats"]["modes"] = [mode]

        modes = []
        modes_weights = []
        chains = []
        chains_weights = []

        if self.multimodes:
            # separate modes. get individual samples for each mode

            # get parameter values and sample probability (=weight) for each mode
            with open(
                os.path.join(
                    self.dir_multinest / f"{self.multinest_prefix}post_separate.dat",
                )
            ) as f:
                lines = f.readlines()
                for idx, line in enumerate(lines):
                    if idx > 2:  # skip the first two lines
                        if lines[idx - 1] == "\n" and lines[idx - 2] == "\n":
                            modes.append(chains)
                            modes_weights.append(chains_weights)
                            chains = []
                            chains_weights = []
                    chain = [float(x) for x in line.split()[2:]]
                    if len(chain) > 0:
                        chains.append(chain)
                        chains_weights.append(float(line.split()[0]))
                modes.append(chains)
                modes_weights.append(chains_weights)
            modes_array = []
            for mode in modes:
                mode_array = np.zeros((len(mode), len(mode[0])))
                for idx, line in enumerate(mode):
                    mode_array[idx, :] = line
                modes_array.append(mode_array)
        else:
            # not running in multimode. Get chains directly from file 1-.txt
            modes_array = [data[:, 2:]]
            chains_weights = [data[:, 0]]
            modes_weights.append(chains_weights[0])
            modes = [0]

        for nmode in range(len(modes)):
            self.debug(f"Nmode: {nmode}")

            mydict = {
                "type": "nest",
                "local_logE": (
                    nest_out["NEST_stats"]["modes"][0]["local log-evidence"],
                    nest_out["NEST_stats"]["modes"][0]["local log-evidence error"],
                ),
                "weights": np.asarray(modes_weights[nmode]),
                "tracedata": modes_array[nmode],
                "fit_params": {},
            }

            nest_stats
            for idx, param_name in enumerate(self.fit_names):
                trace = modes_array[nmode][:, idx]
                q_16, q_50, q_84 = quantile_corner(
                    trace, [0.16, 0.5, 0.84], weights=np.asarray(modes_weights[nmode])
                )

                mydict["fit_params"][param_name] = {
                    "value": q_50,
                    "sigma_m": q_50 - q_16,
                    "sigma_p": q_84 - q_50,
                    "nest_map": nest_stats["modes"][nmode]["maximum a posterior"][idx],
                    "mean": nest_stats["modes"][nmode]["mean"][idx],
                    "nest_sigma": nest_stats["modes"][nmode]["sigma"][idx],
                    "trace": trace,
                }

            nest_out["solutions"][f"solution{nmode}"] = mydict

        return nest_out

    def generate_solution(
        self, output_size=OutputSize.heavy
    ) -> t.Dict[str, AnyValType]:
        """Generate solution."""
        solution = super().generate_solution(output_size=output_size)

        solution["GlobalStats"] = self._multinest_output["NEST_stats"]
        return solution

    def get_samples(self, solution_idx: int) -> npt.NDArray[np.float64]:
        """Get samples from solution."""
        return self._multinest_output["solutions"][f"solution{solution_idx}"][
            "tracedata"
        ]

    def get_weights(self, solution_idx: int) -> npt.NDArray[np.float64]:
        """Get weights from solution."""
        return self._multinest_output["solutions"][f"solution{solution_idx}"]["weights"]

    def get_solution(
        self,
    ) -> t.Generator[
        t.Tuple[
            int,
            npt.NDArray[np.float64],
            npt.NDArray[np.float64],
            t.Tuple[
                t.Tuple[
                    t.Literal["Statistics"],
                    t.Dict[
                        t.Literal["local log-evidence", "local log-evidence error"],
                        float,
                    ],
                ],
                t.Tuple[t.Literal["fit_params"], t.Dict[str, NestFitParam]],
                t.Tuple[t.Literal["tracedata"], npt.NDArray[np.float64]],
                t.Tuple[t.Literal["weights"], npt.NDArray[np.float64]],
            ],
        ],
        None,
        None,
    ]:
        """Get solution as generator."""
        names = self.fit_names
        opt_values = self.fit_values
        opt_map = self.fit_values
        solutions = [
            (k, v)
            for k, v in self._multinest_output["solutions"].items()
            if "solution" in k
        ]
        for k, v in solutions:
            solution_idx = int(k[8:])
            for p_name, p_value in v["fit_params"].items():
                idx = names.index(p_name)
                opt_map[idx] = p_value["nest_map"]
                opt_values[idx] = p_value["value"]

            yield solution_idx, opt_map, opt_values, [
                (
                    "Statistics",
                    {
                        "local log-evidence": self._multinest_output["NEST_stats"][
                            "modes"
                        ][solution_idx]["local log-evidence"],
                        "local log-evidence error": self._multinest_output[
                            "NEST_stats"
                        ]["modes"][solution_idx]["local log-evidence error"],
                    },
                ),
                ("fit_params", v["fit_params"]),
                ("tracedata", v["tracedata"]),
                ("weights", v["weights"]),
            ]

    @classmethod
    def input_keywords(cls) -> t.Tuple[str, ...]:
        """Input keywords for multinest."""
        return (
            "multinest",
            "pymultinest",
        )

    BIBTEX_ENTRIES = [
        r"""
        @article{ refId0,
        author = {{Buchner, J.} and {Georgakakis, A.} and {Nandra, K.} and {Hsu, L.}
        and {Rangel, C.} and {Brightman, M.} and {Merloni, A.} and {Salvato, M.}
        and {Donley, J.} and {Kocevski, D.}},
        title = {X-ray spectral modelling of the AGN obscuring region in the
            CDFS: Bayesian model selection and catalogue},
        DOI= "10.1051/0004-6361/201322971",
        url= "https://doi.org/10.1051/0004-6361/201322971",
        journal = {A\&A},
        year = 2014,
        volume = 564,
        pages = "A125",
        month = "",
        }
        """,
        """
        @ARTICLE{Feroz_multinest,
            author = {{Feroz}, F. and {Hobson}, M.~P. and {Bridges}, M.},
                title = "{MULTINEST: an efficient and robust Bayesian
                inference tool for cosmology and particle physics}",
            journal = {MNRAS},
            keywords = {methods: data analysis, methods: statistical, Astrophysics},
                year = "2009",
                month = "Oct",
            volume = {398},
            number = {4},
                pages = {1601-1614},
                doi = {10.1111/j.1365-2966.2009.14548.x},
        archivePrefix = {arXiv},
            eprint = {0809.3437},
        primaryClass = {astro-ph},
            adsurl = {https://ui.adsabs.harvard.edu/abs/2009MNRAS.398.1601F},
            adsnote = {Provided by the SAO/NASA Astrophysics Data System}
        }
        """,
    ]



"""Base chemical profile class."""

import typing as t

import numpy as np
import numpy.typing as npt

from taurex.cache import GlobalCache, OpacityCache
from taurex.cache.ktablecache import KTableCache
from taurex.data.citation import Citable
from taurex.data.fittable import Fittable, derivedparam
from taurex.log import Logger
from taurex.output import OutputGroup
from taurex.output.writeable import Writeable
from taurex.planet import Planet
from taurex.stellar import Star


class Chemistry(Fittable, Logger, Writeable, Citable):
    """Skeleton for defining chemistry.
    *Abstract Class*

    Must implement methods:

    - :func:`activeGases`
    - :func:`inactiveGases`
    - :func:`activeGasMixProfile`
    - :func:`inactiveGasMixProfile`

    *Active* are those that are actively
    absorbing in the atmosphere. In technical terms they are molecules
    that have absorption cross-sections. You can see which molecules
    are able to actively absorb by doing:
    You can find out what molecules can actively absorb by doing:

            >>> avail_active_mols = OpacityCache().find_list_of_molecules()

    Active gases are only determined at initialization and cannot be changed
    during runtime. If you want to change the active gases you will need to
    reinitialize the chemistry class.


    """

    def __init__(self, name: str):
        """Initialize chemistry.

        Parameters
        ----------
        name : str
            Name of chemistry for logging.

        """
        Logger.__init__(self, name)
        Fittable.__init__(self)

        self.mu_profile = None

        if GlobalCache()["opacity_method"] == "ktables":
            self._avail_active = KTableCache().find_list_of_molecules()
        else:
            self._avail_active = OpacityCache().find_list_of_molecules()
        # self._avail_active = OpacityCache().find_list_of_molecules()
        deactive_list = GlobalCache()["deactive_molecules"]
        if deactive_list is not None:
            self._avail_active = [
                k for k in self._avail_active if k not in deactive_list
            ]

    def set_star_planet(self, star: Star, planet: Planet):
        """Supplies the star and planet to chemistryfor photochemistry reasons.

        Does nothing by default

        Parameters
        ----------

        star: :class:`~taurex.data.stellar.star.Star`
            A star object

        planet: :class:`~taurex.data.planet.Planet`
            A planet object


        """
        pass

    @property
    def availableActive(self) -> t.List[str]:  # noqa: N802
        """Returns a list of available actively absorbing molecules.

        Returns
        -------
        molecules: :obj:`list`
            Actively absorbing molecules
        """
        return self._avail_active

    @property
    def activeGases(self) -> t.List[str]:  # noqa: N802
        """Actively absorbing gases.

        **Requires implementation**

        Should return a list of molecule names

        Returns
        -------
        active : :obj:`list`
            List of active gases

        """
        raise NotImplementedError

    @property
    def inactiveGases(self) -> t.List[str]:  # noqa: N802
        """Non absorbing gases.
        **Requires implementation**

        Should return a list of molecule names

        Returns
        -------
        inactive : :obj:`list`
            List of inactive gases

        """
        raise NotImplementedError

    def initialize_chemistry(
        self,
        nlayers: int,
        temperature_profile: npt.NDArray[np.float64],
        pressure_profile: npt.NDArray[np.float64],
        altitude_profile: t.Optional[npt.NDArray[np.float64]] = None,
    ) -> None:
        """Initialize the profile.

        Parameters
        -----------
        nlayers : int
            number of layers
        temperature_profile : np.ndarray
            temperature profile
        pressure_profile : np.ndarray
            pressure profile
        altitude_profile : np.ndarray , optional
            altitude profile, deprecated

        """
        pass

    @property
    def activeGasMixProfile(self) -> npt.NDArray[np.float64]:  # noqa: N802
        """Mix profile of actively absorbing gases.
        **Requires implementation**

        Should return profiles of shape ``(nactivegases,nlayers)``. Active
        refers to gases that are actively absorbing in the atmosphere.
        Another way to put it these are gases where molecular cross-sections
        are used.

        """

        raise NotImplementedError

    @property
    def inactiveGasMixProfile(self) -> npt.NDArray[np.float64]:  # noqa: N802
        """Mixing profile of non absorbing gases.

        **Requires implementation**

        Should return profiles of shape ``(ninactivegases,nlayers)``.

        """
        raise NotImplementedError

    @property
    def muProfile(self) -> npt.NDArray[np.float64]:  # noqa: N802
        """Molecular weight for each layer of atmosphere in kg

        Returns
        -------
        mix_profile : :obj:`array`

        """
        if self.mu_profile is None:
            self.compute_mu_profile(None)
        return self.mu_profile

    def get_gas_mix_profile(self, gas_name: str) -> npt.NDArray[np.float64]:
        """Returns the mix profile of a particular gas

        Parameters
        ----------
        gas_name : str
            Name of gas

        Returns
        -------
        mixprofile : :obj:`array`
            Mix profile of gas with shape ``(nlayer)``

        """
        if gas_name in self.activeGases:
            idx = self.activeGases.index(gas_name)
            return self.activeGasMixProfile[idx]
        elif gas_name in self.inactiveGases:
            idx = self.inactiveGases.index(gas_name)
            return self.inactiveGasMixProfile[idx]
        else:
            raise KeyError

    def compute_mu_profile(self, nlayers: t.Optional[int] = None) -> None:
        """Computes molecular weight of atmosphere for each layer

        Parameters
        ----------
        nlayers: int
            Number of layers, deprecated
        """

        active = []
        inactive = []

        if self.activeGasMixProfile is not None:
            active = [
                mix * self.get_molecular_mass(gasname)
                for mix, gasname in zip(self.activeGasMixProfile, self.activeGases)
            ]

        if self.inactiveGasMixProfile is not None:
            inactive = [
                mix * self.get_molecular_mass(gasname)
                for mix, gasname in zip(self.inactiveGasMixProfile, self.inactiveGases)
            ]

        total = active + inactive
        if not total:
            raise ValueError("No gases or chemical profile in atmosphere")

        self.mu_profile = np.sum(total, axis=0)

    @property
    def gases(self) -> t.List[str]:
        """Total list of gases in atmosphere"""
        return self.activeGases + self.inactiveGases

    @property
    def mixProfile(self) -> npt.NDArray[np.float64]:  # noqa: N802
        return np.concatenate((self.activeGasMixProfile, self.inactiveGasMixProfile))

    @derivedparam(param_name="mu", param_latex=r"$\mu$", compute=True)
    def mu(self) -> float:
        """Mean molecular weight at surface (amu)."""
        from taurex.constants import AMU

        return self.muProfile[0] / AMU

    def write(self, output: OutputGroup) -> OutputGroup:
        """Writes chemistry class and arguments to file.

        Parameters
        ----------
        output: :class:`~taurex.output.output.Output`

        """
        gas_entry = output.create_group("Chemistry")
        gas_entry.write_string("chemistry_type", self.__class__.__name__)
        gas_entry.write_string_array("active_gases", self.activeGases)
        gas_entry.write_string_array("inactive_gases", self.inactiveGases)
        if self.hasCondensates:
            gas_entry.write_string_array("condensates", self.condensates)
        return gas_entry

    @property
    def condensates(self) -> t.List[str]:
        """
        Returns a list of condensates in the atmosphere.

        Returns
        -------
        active : :obj:`list`
            List of condensates

        """

        return []

    @property
    def hasCondensates(self) -> bool:  # noqa: N802
        """Returns True if there are condensates in the atmosphere."""
        return len(self.condensates) > 0

    @property
    def condensateMixProfile(self) -> npt.NDArray[np.float64]:  # noqa: N802
        """Get condensate mix profile.
        **Requires implementation**

        Should return profiles of shape ``(ncondensates,nlayers)``.
        """
        if len(self.condensates) == 0:
            return None
        else:
            raise NotImplementedError

    def get_condensate_mix_profile(
        self, condensate_name: str
    ) -> npt.NDArray[np.float64]:
        """Returns the mix profile of a particular condensate.

        Parameters
        ----------
        condensate_name : str
            Name of condensate

        Returns
        -------
        mixprofile : :obj:`array`
            Mix profile of condensate with shape ``(nlayer)``

        """
        if condensate_name in self.condensates:
            index = self.condensates.index(condensate_name)
            return self.condensateMixProfile[index]
        else:
            raise KeyError(f"Condensate {condensate_name} not found in chemistry")

    def get_molecular_mass(self, molecule: str) -> float:
        """Returns the molecular mass of a molecule."""
        from taurex.util import get_molecular_weight

        return get_molecular_weight(molecule)

    @classmethod
    def input_keywords(cls) -> t.Tuple[str, ...]:
        """Input keywords for chemistry class."""
        raise NotImplementedError



"""Module for loading in PHOENIX spectra."""
import math
import os
import typing as t

import numpy as np
import numpy.typing as npt

from taurex.cache import GlobalCache
from taurex.constants import MSOL
from taurex.output import OutputGroup
from taurex.types import PathLike

from .star import BlackbodyStar


class PhoenixStar(BlackbodyStar):
    """A star that uses the PHOENIX synthetic stellar atmosphere spectrums.

    These spectrums are read from ``.gits.gz`` files in a directory given by
    ``phoenix_path``
    Each file must contain the spectrum for one temperature

    Parameters
    ----------

    phoenix_path: str, **required**
        Path to folder containing phoenix ``fits.gz`` files

    temperature: float, optional
        Stellar temperature in Kelvin

    radius: float, optional
        Stellar radius in Solar radius

    metallicity: float, optional
        Metallicity in solar values

    mass: float, optional
        Stellar mass in solar mass

    distance: float, optional
        Distance from Earth in pc

    magnitudeK: float, optional
        Maginitude in K band


    Raises
    ------
    Exception
        Raised when no phoenix path is defined


    """

    def __init__(
        self,
        temperature: t.Optional[float] = 5000,
        radius: t.Optional[float] = 1.0,
        metallicity: t.Optional[float] = 1.0,
        mass: t.Optional[float] = 1.0,
        distance: t.Optional[float] = 1,
        magnitudeK: t.Optional[float] = 10.0,  # noqa: N803
        phoenix_path: t.Optional[PathLike] = None,
        retro_version_file: t.Optional[PathLike] = None,
    ):
        super().__init__(
            temperature=temperature,
            radius=radius,
            distance=distance,
            magnitudeK=magnitudeK,
            mass=mass,
            metallicity=metallicity,
        )
        self._phoenix_path = phoenix_path

        self.retro_version_file = retro_version_file
        # CAN BE OBTAINED FROM:
        # ftp://phoenix.astro.physik.uni-goettingen.de/HiResFITS/PHOENIX-ACES-AGSS-COND-2011/

        if self._phoenix_path is None or not os.path.isdir(self._phoenix_path):
            self._phoenix_path = GlobalCache()["phoenix_path"]

        if self._phoenix_path is None or not os.path.isdir(self._phoenix_path):
            self.error(
                "No file path or incorrect path to "
                f"phoenix files defined - {self._phoenix_path}"
            )
            raise Exception(
                "No file path or incorrect path "
                f"to phoenix files defined - {self._phoenix_path}"
            )

        self.info("Star is PHOENIX type")

        self.get_avail_phoenix()
        self.use_blackbody = False
        self.recompute_spectra()
        # self.preload_phoenix_spectra()

    def compute_logg(self) -> float:
        """Computes log(surface_G)."""
        import astropy.units as u
        from astropy.constants import G

        mass = self._mass * u.kg
        radius = self._radius * u.m

        small_g = (G * mass) / (radius**2)

        small_g = small_g.to(u.cm / u.s**2)

        return math.log10(small_g.value)

    def recompute_spectra(self) -> None:
        """Recompute spectra as needed."""
        if (
            self.temperature > self._T_list.max()
            or self.temperature < self._T_list.min()
        ):
            self._logg = self.compute_logg()
            self.use_blackbody = True
        else:
            self.use_blackbody = False
            self._logg = self.compute_logg()
            f = self.find_nearest_file()
            self.read_spectra(f)

    def read_spectra(self, p_file: PathLike) -> None:
        """Reads in the spectra from a given file."""
        import astropy.units as u
        from astropy.io import fits

        if self.retro_version_file is not None:
            with fits.open(p_file) as hdu:
                with fits.open(self.retro_version_file) as hdu2:
                    str_unit = hdu2[0].header["UNIT"]
                    wl = hdu2[0].data * u.Unit(str_unit)

                    str_unit = hdu[0].header["BUNIT"]
                    sed = hdu[0].data * u.Unit(str_unit)

                    self.wngrid = 10000 / (wl.to(u.micron).value)
                    argidx = np.argsort(self.wngrid)
                    self._base_sed = sed.to(u.W / u.m**2 / u.micron).value
                    self.wngrid = self.wngrid[argidx]
                    self._base_sed = self._base_sed[argidx]
        else:
            with fits.open(p_file) as hdu:
                str_unit = hdu[1].header["TUNIT1"]
                wl = hdu[1].data.field("Wavelength") * u.Unit(str_unit)

                str_unit = hdu[1].header["TUNIT2"]
                sed = hdu[1].data.field("Flux") * u.Unit(str_unit)

                self.wngrid = 10000 / (wl.value)
                argidx = np.argsort(self.wngrid)
                self._base_sed = sed.to(u.W / u.m**2 / u.micron).value
                self.wngrid = self.wngrid[argidx]
                self._base_sed = self._base_sed[argidx]

    @property
    def temperature(self) -> float:
        """Effective Temperature in Kelvin.

        Returns
        -------
        T: float

        """
        return self._temperature

    @temperature.setter
    def temperature(self, value: float) -> None:
        self._temperature = value
        self.recompute_spectra()

    @property
    def mass(self) -> float:
        """Mass of star in solar mass.

        Returns
        -------
        M: float

        """
        return self._mass

    @mass.setter
    def mass(self, value: float) -> None:
        self._mass = value * MSOL
        self.recompute_spectra()

    def find_nearest_file(self) -> str:
        """Finds the nearest file to the current stellar parameters."""
        idx = self._index_finder([self._temperature, self._logg, self._metallicity])
        return self._files[int(idx)]

    def get_avail_phoenix(self) -> None:
        import glob

        from scipy.interpolate import NearestNDInterpolator

        if self.retro_version_file is not None:
            files = glob.glob(os.path.join(self._phoenix_path, "*.fits"))
        else:
            files = glob.glob(os.path.join(self._phoenix_path, "*.spec.fits.gz"))
        # files = glob.glob(os.path.join(self._phoenix_path, '*.fits'))

        self._files = files
        if self.retro_version_file is not None:
            self._T_list = np.array(
                [np.float64(os.path.basename(k)[3:8]) for k in files]
            )
            self._Logg_list = np.array(
                [np.float64(os.path.basename(k)[9:13]) for k in files]
            )
            self._Z_list = np.array(
                [np.float64(os.path.basename(k)[14:17]) for k in files]
            )
        else:
            self._T_list = (
                np.array([np.float64(os.path.basename(k)[3:8]) for k in files]) * 100
            )
            self._Logg_list = np.array(
                [np.float64(os.path.basename(k)[8:12]) for k in files]
            )
            self._Z_list = np.array(
                [np.float64(os.path.basename(k)[13:16]) for k in files]
            )
        self._index_finder = NearestNDInterpolator(
            (self._T_list, self._Logg_list, self._Z_list),
            np.arange(0, self._T_list.shape[0]),
            rescale=True,
        )

    def initialize(self, wngrid: npt.NDArray[np.float64]):
        """Initializes and interpolates the spectral emission density

        SED is interpolated to the current stellar temperature and
        given wavenumber grid

        Parameters
        ----------
        wngrid:
            Wavenumber grid to interpolate the SED to

        """
        # If temperature outside of range, use blavkbody
        if self.use_blackbody:
            self.warning(
                "Using black body as temperature is "
                "outside of Star temeprature range %s",
                self.temperature,
            )
            super().initialize(wngrid)
        else:
            sed = self._base_sed
            self.sed = np.interp(wngrid, self.wngrid, sed)

    @property
    def spectralEmissionDensity(self) -> float:  # noqa: N802
        """Spectral emmision density in W/m^2/um.

        Returns
        -------
        sed: :obj:`array`
        """
        return self.sed

    def write(self, output: OutputGroup) -> OutputGroup:
        """Write to output group."""
        star = super().write(output)
        star.write_string("phoenix_path", self._phoenix_path)
        return star

    @classmethod
    def input_keywords(cls) -> t.Tuple[str, ...]:
        return ("phoenix",)

    BIBTEX_ENTRIES = [
        r"""
        @article{ refId0,
            author = {{Husser, T.-O.} and {Wende-von Berg, S.} and
            {Dreizler, S.} and {Homeier, D.} and {Reiners, A.} and
            {Barman, T.} and {Hauschildt, P. H.}},
            title = {A new extensive library of PHOENIX stellar atmospheres and
                synthetic spectra},
            DOI= "10.1051/0004-6361/201219058",
            url= "https://doi.org/10.1051/0004-6361/201219058",
            journal = {A\&A},
            year = 2013,
            volume = 553,
            pages = "A6",
            month = "",
        }
        """
    ]



"""Spectra loaded from an array."""
import typing as t

import numpy as np
import numpy.typing as npt

from taurex.util import wnwidth_to_wlwidth

from .spectrum import BaseSpectrum


class ArraySpectrum(BaseSpectrum):
    """Loads an observed spectrum from an array

    Loads spectrum and computes bin
    edges and bin widths. Spectrum shape(nbins, 3-4) with 3-4 columns with
    ordering:
        1. wavelength (um)
        2. spectral data
        3. error
        4. (optional) bin width

    If no bin width is present then they are computed.


    """

    def __init__(self, spectrum: t.Optional[npt.NDArray[np.float64]] = None):
        """Initialize.


        Parameters
        -----------
        spectrum:
            Array of shape (nbins, 3-4) with 3-4 columns, with ordering:
            1. wavelength (um)
            2. spectral data
            3. error
            4. (optional) bin width


        """
        super().__init__(self.__class__.__name__)

        self._obs_spectrum = spectrum
        self._bin_widths = None
        self._bin_edges = None

        self._sort_spectrum()
        self._process_spectrum()

        self._wnwidths = wnwidth_to_wlwidth(self.wavelengthGrid, self._bin_widths)

    def _sort_spectrum(self) -> None:
        """Sorts the spectrum by wavelength"""
        self._obs_spectrum = self._obs_spectrum[
            self._obs_spectrum[:, 0].argsort(axis=0)[::-1]
        ]

    def _process_spectrum(self) -> None:
        """Seperate out the observed data, error, grid and binwidths

        If bin widths are not present then they are
        calculated here
        """
        if self.rawData.shape[1] == 4:
            self._bin_widths = self._obs_spectrum[:, 3]
            obs_wl = self.wavelengthGrid[::-1]
            obs_bw = self._bin_widths[::-1]

            bin_edges = np.zeros(shape=(len(self._bin_widths) * 2,))

            bin_edges[0::2] = obs_wl - obs_bw / 2
            bin_edges[1::2] = obs_wl + obs_bw / 2
            # bin_edges[-1] = obs_wl[-1]-obs_bw[-1]/2.

            self._bin_edges = bin_edges[::-1]
        else:
            self.manual_binning()

    @property
    def rawData(self) -> npt.NDArray[np.float64]:  # noqa: N802
        """Data read from file."""
        return self._obs_spectrum

    @property
    def spectrum(self) -> npt.NDArray[np.float64]:
        """The spectrum itself."""
        return self._obs_spectrum[:, 1]

    @property
    def wavelengthGrid(self) -> npt.NDArray[np.float64]:  # noqa: N802
        """Wavelength grid in microns."""
        return self.rawData[:, 0]

    @property
    def wavenumberGrid(self) -> npt.NDArray[np.float64]:  # noqa: N802
        """Wavenumber grid in cm-1"""
        return 10000 / self.wavelengthGrid

    @property
    def binEdges(self) -> npt.NDArray[np.float64]:  # noqa: N802
        """Bin edges."""
        return 10000 / self._bin_edges

    @property
    def binWidths(self) -> npt.NDArray[np.float64]:  # noqa: N802
        """bin widths."""
        return self._wnwidths

    @property
    def errorBar(self) -> npt.NDArray[np.float64]:  # noqa: N802
        """Error bars for the spectrum."""
        return self.rawData[:, 2]

    def manual_binning(self) -> None:
        """Performs the calculation of bin edges when none are present."""
        from taurex.util import compute_bin_edges

        self._bin_edges, self._bin_widths = compute_bin_edges(self.wavelengthGrid)

    @classmethod
    def input_keywords(cls) -> t.Tuple[str, ...]:
        """Input keywords for this class."""
        return ("array",)



"""Functions related to computing emission spectrums."""

import math
import typing as t

import numpy as np
import numpy.typing as npt

from taurex.constants import KBOLTZ, PI, PLANCK, SPDLIGT

# NUMBA Functions
try:
    import numba
    from numba import float64

    @numba.vectorize([float64(float64)], fastmath=True)
    def _convert_lamb(lamb: npt.NDArray[np.float64]) -> npt:
        """Convert wavenumber in :math:`\\mu m` to :math:`m`."""
        return 10000 * 1e-6 / lamb

    @numba.vectorize([float64(float64, float64)], fastmath=True)
    def _black_body_vec(wl: npt.NDArray[np.float64], temp: float):
        """Compute black body spectrum."""
        return (
            (PI * (2.0 * PLANCK * SPDLIGT**2) / (wl) ** 5)
            * (1.0 / (np.exp((PLANCK * SPDLIGT) / (wl * KBOLTZ * temp)) - 1))
            * 1e-6
        )

    @numba.njit(fastmath=True, parallel=False)
    def black_body_numba(lamb: npt.NDArray[np.float64], temp: float):
        """Compute black body spectrum using numba."""
        wl = _convert_lamb(lamb)
        return _black_body_vec(wl, temp)

    @numba.njit(fastmath=True, parallel=False)
    def black_body_numba_II(lamb, temp):  # noqa: N802
        """Compute black body spectrum (alt algo) using numba."""
        n = lamb.shape[0]
        out = np.zeros_like(lamb)
        conversion = 10000 * 1e-6
        # for n in range(N):
        #     wl[n] = 10000*1e-6/lamb[n]

        factor = PI * (2.0 * PLANCK * SPDLIGT**2) * 1e-6 / conversion**5
        c2 = PLANCK * SPDLIGT / (KBOLTZ * temp) / conversion

        for x in range(n):
            out[x] = factor * lamb[x] ** 5 / (math.exp(c2 * lamb[x]) - 1)

        return out

except ImportError:
    print("Numba not installed, using numpy instead")

    def black_body_numba(lamb: npt.NDArray[np.float64], temp: float):
        """Compute black body spectrum using numpy (numba not available)."""
        return black_body_numpy(lamb, temp)

    def black_body_numba_II(lamb: npt.NDArray[np.float64], temp: float):  # noqa: N802
        """Compute black body spectrum using numpy (numba not available)."""
        return black_body_numpy(lamb, temp)


def black_body_numexpr(lamb: npt.NDArray, temp: npt.NDArray) -> npt.NDArray:
    """Compute black body spectrum using numexpr."""
    import numexpr as ne

    wl = ne.evaluate("10000*1e-6/lamb")  # noqa: F841

    return ne.evaluate(
        "(PI* (2.0*PLANCK*SPDLIGT**2)/(wl)**5) * (1.0/(exp((PLANCK * SPDLIGT)"
        " / (wl * KBOLTZ * temp))-1))*1e-6"
    )


def black_body_numpy(lamb: npt.NDArray, temp: npt.NDArray) -> npt.NDArray:
    """Compute black body spectrum using numpy.

    Parameters
    ----------
    lamb : npt.NDArray
        Wavelengths in microns
    temp : npt.NDArray
        Temperature in Kelvin

    Returns
    -------
    npt.NDArray
        Black body spectrum

    """
    h = 6.62606957e-34
    c = 299792458
    k = 1.3806488e-23
    pi = np.pi

    temp = np.atleast_1d(temp)

    lamb = np.atleast_1d(lamb)

    temp_ = temp.ravel()
    lamb_ = lamb.ravel()

    final_shape = temp_.shape + lamb_.shape

    temp_ = np.broadcast_to(temp_[:, None], final_shape)
    lamb_ = np.broadcast_to(lamb_, final_shape)
    wl = 10000 / lamb_
    exponent = np.exp((h * c) / (wl * 1e-6 * k * temp_))
    bb = (pi * (2.0 * h * c**2) / (wl * 1e-6) ** 5) * (1.0 / (exponent - 1))
    final = bb * 1e-6

    if temp.size == 1:
        final = final.squeeze(axis=0)
    else:
        final = final.reshape(temp.shape + (-1,))
    if lamb.size == 1:
        final = final.squeeze(axis=-1)
    elif final.ndim > 1:
        final = final.reshape(final.shape[:-1] + lamb.shape)

    return final


def integrate_emission_layer(
    dtau: npt.NDArray, layer_tau: npt.NDArray, mu: npt.NDArray, bb: npt.NDArray
) -> t.Tuple[npt.NDArray, npt.NDArray]:
    """Integrate emission layer.

    Parameters
    ----------
    dtau : npt.NDArray
        Optical depth of layer
    layer_tau : npt.NDArray
        Optical depth of layer
    mu : npt.NDArray
        Cosine of zenith angle
    BB : npt.NDArray
        Black body spectrum

    Returns
    -------
    t.Tuple[npt.NDArray, npt.NDArray]
        Integrated emission layer, optical depth of layer

    """

    _mu = 1 / mu[:, None]
    _tau = np.exp(-layer_tau) - np.exp(-dtau)

    return bb * (np.exp(-layer_tau * _mu) - np.exp(-dtau * _mu)), _tau


black_body = black_body_numba
"""Black body function to use."""



"""Retrieval using nestle library."""
import time
import typing as t

import nestle
import numpy as np
import numpy.typing as npt

from taurex.model import ForwardModel
from taurex.output import OutputGroup
from taurex.spectrum import BaseSpectrum
from taurex.util import quantile_corner, recursively_save_dict_contents_to_output

from .optimizer import FitParamOutput, Optimizer


class NestleStatsOutput(t.TypedDict):
    """Dictionary for storing nestle stats output."""

    LogEvidence: float
    LogEvidenceError: float
    Peakiness: float


class NestleSolutionOutput(t.TypedDict):
    """Dictionary for storing nestle solution output."""

    samples: npt.NDArray[np.float64]
    weights: npt.NDArray[np.float64]
    covariance: npt.NDArray[np.float64]
    fitparams: t.Dict[str, FitParamOutput]


class NestleOptimizer(Optimizer):
    """An optimizer that uses the `nestle <http://kylebarbary.com/nestle/>`_ library
    to perform optimization.

    """

    def __init__(
        self,
        observed: t.Optional[BaseSpectrum] = None,
        model: t.Optional[ForwardModel] = None,
        num_live_points: t.Optional[int] = 1500,
        method: t.Optional[t.Literal["single", "multi", "mcmc"]] = "multi",
        tol: t.Optional[float] = 0.5,
        sigma_fraction: t.Optional[float] = 0.1,
    ):
        """Initialize and setup nestle.


        Parameters
        ----------
        observed:
            Observed spectrum to optimize to

        model:
            Forward model to optimize

        num_live_points:
            Number of live points to use in sampling

        method:
            Nested sampling method to use. ``classic`` uses MCMC exploration,
            ``single`` uses a single ellipsoid and ``multi`` uses
            multiple ellipsoids (similar to Multinest)

        tol:
            Evidence tolerance value to stop the fit. This is based on
            an estimate of the remaining prior volumes
            contribution to the evidence.

        sigma_fraction:
            Fraction of weights to use in computing the error. (Default: 0.1)

        """
        super().__init__("Nestle", observed, model, sigma_fraction)
        self._nlive = int(num_live_points)  # number of live points
        self._method = method  # use MutliNest algorithm

        self._tol = tol  # the stopping criterion
        self._nestle_output: NestleSolutionOutput = None

    @property
    def tolerance(self) -> float:
        """Tolerance value for stopping the fit."""
        return self._tol

    @tolerance.setter
    def tolerance(self, value: float) -> None:
        """Set the tolerance value for stopping the fit."""
        self._tol = value

    @property
    def numLivePoints(self) -> int:  # noqa: N802
        """Number of live points to use in the fit."""
        return self._nlive

    @numLivePoints.setter
    def numLivePoints(self, value: int) -> None:  # noqa: N802
        """Set the number of live points to use in the fit."""
        self._nlive = value

    def compute_fit(self) -> None:
        """Computes the fit using nestle."""

        def nestle_uniform_prior(theta):
            # prior distributions called by multinest. Implements a uniform prior
            # converting parameters from normalised grid to uniform prior

            return tuple(self.prior_transform(theta))

        ndim = len(self.fitting_parameters)
        self.warning("Beginning fit......")
        ndims = ndim  # two parameters

        t0 = time.time()

        res = nestle.sample(
            self.log_likelihood,
            nestle_uniform_prior,
            ndims,
            method="multi",
            npoints=self.numLivePoints,
            dlogz=self.tolerance,
            callback=nestle.print_progress,
        )

        res = t.cast(nestle.Result, res)
        t1 = time.time()

        timenestle = t1 - t0

        print(res.summary())

        self.warning("Time taken to run 'Nestle' is %s seconds", timenestle)

        self.warning("Fit complete.....")

        self._nestle_output = self.store_nestle_output(res)

    def get_samples(self, solution_idx: int) -> npt.NDArray[np.float64]:
        """Returns the samples from the fit."""
        return self._nestle_output["solution"]["samples"]

    def get_weights(self, solution_idx: int) -> npt.NDArray[np.float64]:
        """Returns the weights of the samples."""
        return self._nestle_output["solution"]["weights"]

    def get_solution(
        self,
    ) -> t.Generator[
        t.Tuple[
            int,
            npt.NDArray[np.float64],
            npt.NDArray[np.float64],
            t.Tuple[
                t.Tuple[t.Literal["Statistics"], float],
                t.Tuple[t.Literal["fit_params"], t.Dict[str, FitParamOutput]],
                t.Tuple[t.Literal["tracedata"], npt.NDArray[np.float64]],
                t.Tuple[t.Literal["weights"], npt.NDArray[np.float64]],
            ],
        ],
        None,
        None,
    ]:
        """Generator for solutions and their median and MAP values

        Yields
        ------

        solution_no:
            Solution number (always 0)

        map:
            Map values

        median:
            Median values

        extra:
            statistics, fit_params, tracedata, weights

        """

        names = self.fit_names
        opt_map = self.fit_values
        opt_values = self.fit_values
        for k, v in self._nestle_output["solution"]["fitparams"].items():
            # if k.endswith('_derived'):
            #     continue
            idx = names.index(k)
            opt_map[idx] = v["map"]
            opt_values[idx] = v["value"]

        yield 0, opt_map, opt_values, (
            ("Statistics", self._nestle_output["Stats"]),
            ("fit_params", self._nestle_output["solution"]["fitparams"]),
            ("tracedata", self._nestle_output["solution"]["samples"]),
            ("weights", self._nestle_output["solution"]["weights"]),
        )

    def write_optimizer(self, output: OutputGroup) -> OutputGroup:
        """Writes the optimizer to the output group."""
        opt = super().write_optimizer(output)

        # number of live points
        opt.write_scalar("num_live_points", self._nlive)
        # maximum no. of iterations (0=inf)
        opt.write_string("method", self._method)
        # search for multiple modes
        opt.write_scalar("tol", self._tol)

        return opt

    def write_fit(self, output: OutputGroup) -> OutputGroup:
        """Writes the fit to the output group."""
        fit = super().write_fit(output)

        if self._nestle_output:
            recursively_save_dict_contents_to_output(output, self._nestle_output)

        return fit

    def store_nestle_output(self, result: nestle.Result) -> NestleSolutionOutput:
        """This turns the output fron nestle into a dictionary

        Contains summary statistics and the solution.

        """
        nestle_output = {}
        nestle_output["Stats"] = {}
        nestle_output["Stats"]["Log-Evidence"] = result.logz
        nestle_output["Stats"]["Log-Evidence-Error"] = result.logzerr
        nestle_output["Stats"]["Peakiness"] = result.h

        fit_param = self.fit_names

        samples = result.samples
        weights = result.weights

        mean, cov = nestle.mean_and_cov(samples, weights)
        nestle_output["solution"] = {}
        nestle_output["solution"]["samples"] = samples
        nestle_output["solution"]["weights"] = weights
        nestle_output["solution"]["covariance"] = cov
        nestle_output["solution"]["fitparams"] = {}

        max_weight = weights.argmax()

        table_data = []

        for idx, param_name in enumerate(fit_param):
            param = {}
            param["mean"] = mean[idx]
            param["sigma"] = cov[idx]
            trace = samples[:, idx]
            q_16, q_50, q_84 = quantile_corner(
                trace, [0.16, 0.5, 0.84], weights=np.asarray(weights)
            )
            param["value"] = q_50
            param["sigma_m"] = q_50 - q_16
            param["sigma_p"] = q_84 - q_50
            param["trace"] = trace
            param["map"] = trace[max_weight]
            table_data.append((param_name, q_50, q_50 - q_16))

            nestle_output["solution"]["fitparams"][param_name] = param

        return nestle_output

    @classmethod
    def input_keywords(cls) -> t.Tuple[str, ...]:
        return ("nestle",)

    BIBTEX_ENTRIES = [
        """@misc{nestle,

        author = {Kyle Barbary},
        title = {Nestle sampling library},
        publisher = {GitHub},
        journal = {GitHub repository},
        year = 2015,
        howpublished = {https://github.com/kbarbary/nestle},
        }"""
    ]



"""Module for NPoint temperature profile."""

import typing as t

import numpy as np
import numpy.typing as npt

from taurex.data.fittable import fitparam
from taurex.exceptions import InvalidModelException
from taurex.output import OutputGroup
from taurex.util import movingaverage

from .tprofile import TemperatureProfile


class InvalidTemperatureException(InvalidModelException):
    """Exception that is called when temperature profile is invalid."""

    pass


class NPoint(TemperatureProfile):
    """Temperature profile defined and smoothed by user points.

    A temperature profile that is defined at various heights of the
    atmopshere and then smoothend.

    At minimum, temepratures on both the top ``T_top`` and surface
    ``T_surface`` must be defined.
    If any intermediate points are given as ``temperature_points``
    then the same number of ``pressure_points``
    must be given as well.

    A 2-point temperature profile has ``len(temperature_points) == 0``
    A 3-point temperature profile has ``len(temperature_points) == 1``

    etc.
    """

    def __init__(
        self,
        T_surface: t.Optional[float] = 1500.0,  # noqa: N803
        T_top: t.Optional[float] = 200.0,  # noqa: N803
        P_surface: t.Optional[float] = None,  # noqa: N803
        P_top: t.Optional[float] = None,  # noqa: N803
        temperature_points: t.Optional[npt.ArrayLike] = None,
        pressure_points: t.Optional[npt.ArrayLike] = None,
        smoothing_window: t.Optional[int] = 10,
        limit_slope: t.Optional[int] = 9999999,
    ):
        """Initialize NPoint temperature profile.

        Parameters
        ----------
        T_surface : float
            BOA temperature in Kelvin
        T_top : float
            TOA temperature in Kelvin
        P_surface : float
            BOA pressure in Pa
        P_top : float
            TOA pressure in Pa
        temperature_points : array-like
            Temperature points
        pressure_points : array-like
            Pressure points
        smoothing_window : int
            Smoothing window
        limit_slope : int
            Gradient limit to be considered valid
        """
        temperature_points = (
            temperature_points if temperature_points is not None else []
        )
        pressure_points = pressure_points if pressure_points is not None else []

        super().__init__(f"{len(temperature_points) + 2}Point")

        if not hasattr(temperature_points, "__len__"):
            raise ValueError("t_point is not an iterable")

        if len(temperature_points) != len(pressure_points):
            self.error("Number of temeprature points != number of " "pressure points")
            self.error(
                "len(t_points) = %s /= " "len(p_points) = %s",
                len(temperature_points),
                len(pressure_points),
            )
            raise ValueError("Incorrect_number of temp and pressure points")

        self.info("Npoint temeprature profile is initialized")
        self.debug("Passed temeprature points %s", temperature_points)
        self.debug("Passed pressure points %s", pressure_points)
        self._t_points = temperature_points
        self._p_points = pressure_points
        self._T_surface = T_surface
        self._T_top = T_top
        self._P_surface = P_surface
        self._P_top = P_top
        self._smooth_window = smoothing_window
        self._limit_slope = limit_slope
        self.generate_pressure_fitting_params()
        self.generate_temperature_fitting_params()

    @fitparam(
        param_name="T_surface",
        param_latex=r"$T_\mathrm{surf}$",
        default_fit=False,
        default_bounds=[300, 2500],
    )
    def temperatureSurface(self) -> float:  # noqa: N802
        """Temperature at planet surface in Kelvin"""
        return self._T_surface

    @temperatureSurface.setter
    def temperatureSurface(self, value: float) -> None:  # noqa: N802
        """Temperature at planet surface in Kelvin"""
        self._T_surface = value

    @fitparam(
        param_name="T_top",
        param_latex=r"$T_\mathrm{top}$",
        default_fit=False,
        default_bounds=[300, 2500],
    )
    def temperatureTop(self) -> float:  # noqa: N802
        """Temperature at top of atmosphere in Kelvin"""
        return self._T_top

    @temperatureTop.setter
    def temperatureTop(self, value: float) -> None:  # noqa: N802
        """Temperature at top of atmosphere in Kelvin"""
        self._T_top = value

    @fitparam(
        param_name="P_surface",
        param_latex=r"$P_\mathrm{surf}$",
        default_fit=False,
        default_bounds=[1e3, 1e2],
        default_mode="log",
    )
    def pressureSurface(self) -> float:  # noqa: N802
        """Pressure at planet surface in Pa"""
        return self._P_surface

    @pressureSurface.setter
    def pressureSurface(self, value: float) -> None:  # noqa: N802
        """Pressure at planet surface in Pa"""
        self._P_surface = value

    @fitparam(
        param_name="P_top",
        param_latex=r"$P_\mathrm{top}$",
        default_fit=False,
        default_bounds=[1e-5, 1e-4],
        default_mode="log",
    )
    def pressureTop(self) -> float:  # noqa: N802
        """Pressure at top of atmosphere in Pa"""
        return self._P_top

    @pressureTop.setter
    def pressureTop(self, value: float) -> None:  # noqa: N802
        """Pressure at top of atmosphere in Pa"""
        self._P_top = value

    def generate_pressure_fitting_params(self) -> None:
        """Generates the fitting parameters for the pressure points.

        These are given the name ``P_point(number)`` for example, if two extra
        pressure points are defined between the top and surface then the
        fitting parameters generated are ``P_point0`` and ``P_point1``
        """

        bounds = [1e5, 1e3]
        for idx, _ in enumerate(self._p_points):
            point_num = idx + 1
            param_name = f"P_point{point_num}"
            param_latex = f"$P_{point_num}$"

            def read_point(self, idx=idx):
                return self._p_points[idx]

            def write_point(self, value, idx=idx):
                self._p_points[idx] = value

            read_point.__doc__ = f"Pressure point {point_num} in Pa"

            fget_point = read_point
            fset_point = write_point
            self.debug("FGet_location %s", fget_point)
            default_fit = False
            self.add_fittable_param(
                param_name,
                param_latex,
                fget_point,
                fset_point,
                "log",
                default_fit,
                bounds,
            )

    def generate_temperature_fitting_params(self) -> None:
        """Generates the fitting parameters for the temperature points.

        These are given the name ``T_point(number)`` for example, if two extra
        temeprature points are defined between the top and surface then the
        fitting parameters generated are ``T_point0`` and ``T_point1``
        """

        bounds = [300, 2500]
        for idx, _ in enumerate(self._t_points):
            point_num = idx + 1
            param_name = f"T_point{point_num}"
            param_latex = f"$T_{point_num}$"

            def read_point(self, idx=idx):
                return self._t_points[idx]

            def write_point(self, value, idx=idx):
                self._t_points[idx] = value

            read_point.__doc__ = f"Temperature point {point_num} in K"

            fget_point = read_point
            fset_point = write_point
            self.debug("FGet_location %s %s", fget_point, fget_point(self))
            default_fit = False
            self.add_fittable_param(
                param_name,
                param_latex,
                fget_point,
                fset_point,
                "linear",
                default_fit,
                bounds,
            )

    def check_profile(self, ppt: npt.ArrayLike, tpt: npt.ArrayLike):
        """Checks the validity of the temperature profile.

        Parameters
        ----------
        ppt : array-like
            Pressure points
        tpt : array-like
            Temperature points

        Raises
        ------
        InvalidTemperatureException
            If the temperature profile is invalid


        """
        if any(ppt[i] <= ppt[i + 1] for i in range(len(ppt) - 1)):
            self.warning(
                "Temperature profile is not valid - a pressure point is inverted"
            )
            raise InvalidTemperatureException

        if any(
            abs((tpt[i + 1] - tpt[i]) / (np.log10(ppt[i + 1]) - np.log10(ppt[i])))
            >= self._limit_slope
            for i in range(len(ppt) - 1)
        ):
            self.warning("Temperature profile is not valid - profile slope too high")
            raise InvalidTemperatureException

    @property
    def profile(self) -> npt.NDArray[np.float64]:
        """Returns a smoothed temperature profile.

        Computes the temperature profile from the given points and then
        smooths it using a moving average.

        """
        t_nodes = [self._T_surface, *self._t_points, self._T_top]

        p_surface = self._P_surface
        if p_surface is None or p_surface < 0:
            p_surface = self.pressure_profile[0]

        p_top = self._P_top
        if p_top is None or p_top < 0:
            p_top = self.pressure_profile[-1]

        p_nodes = [p_surface, *self._p_points, p_top]

        self.check_profile(p_nodes, t_nodes)

        smooth_window = self._smooth_window

        if np.all(t_nodes == t_nodes[0]):
            return np.ones_like(self.pressure_profile) * t_nodes[0]

        tp = np.interp(
            (np.log10(self.pressure_profile[::-1])),
            np.log10(p_nodes[::-1]),
            t_nodes[::-1],
        )

        # smoothing T-P profile
        wsize = int(self.nlayers * (smooth_window / 100.0))
        if wsize % 2 == 0:
            wsize += 1
        tp_smooth = movingaverage(tp, wsize)
        border = np.int64((len(tp) - len(tp_smooth)) / 2)

        foo = tp[::-1]
        if len(tp_smooth) == len(foo):
            foo = tp_smooth[::-1]
        else:
            foo[border:-border] = tp_smooth[::-1]

        return foo

    def write(self, output: OutputGroup) -> OutputGroup:
        """Write NPoint temperature profile to output group."""
        temperature = super().write(output)

        temperature.write_scalar("T_surface", self._T_surface)
        temperature.write_scalar("T_top", self._T_top)
        temperature.write_array("temperature_points", np.array(self._t_points))

        p_surface = self._P_surface
        p_top = self._P_top
        if not p_surface:
            p_surface = -1
        if not p_top:
            p_top = -1

        temperature.write_scalar("P_surface", p_surface)
        temperature.write_scalar("P_top", p_top)
        temperature.write_array("pressure_points", np.array(self._p_points))

        temperature.write_scalar("smoothing_window", self._smooth_window)

        return temperature

    @classmethod
    def input_keywords(cls) -> t.Tuple[str, ...]:
        """Return all input keywords."""
        return ("npoint",)



"""Module for mixin classes."""

from .core import (
    AnyMixin,
    ChemistryMixin,
    ContributionMixin,
    ForwardModelMixin,
    GasMixin,
    InstrumentMixin,
    Mixin,
    ObservationMixin,
    OptimizerMixin,
    PlanetMixin,
    PressureMixin,
    SpectrumMixin,
    StarMixin,
    TemperatureMixin,
    enhance_class,
    find_mapped_mixin,
)
from .mixins import MakeFreeMixin, TempScaler

__all__ = [
    "Mixin",
    "StarMixin",
    "TemperatureMixin",
    "PlanetMixin",
    "ContributionMixin",
    "ChemistryMixin",
    "PressureMixin",
    "ForwardModelMixin",
    "SpectrumMixin",
    "OptimizerMixin",
    "ObservationMixin",
    "GasMixin",
    "InstrumentMixin",
    "enhance_class",
    "MakeFreeMixin",
    "AnyMixin",
    "TempScaler",
    "find_mapped_mixin",
]



"""Pressure profile from array."""
import typing as t

import numpy as np
import numpy.typing as npt

from taurex.output import OutputGroup

from .pressureprofile import PressureProfile


class ArrayPressureProfile(PressureProfile):
    """Pressure profile from an array for each layer"""

    def __init__(
        self,
        array: npt.NDArray[np.float64],
        reverse: t.Optional[bool] = False,
        is_central: t.Optional[bool] = True,
    ):
        """Initialize the pressure profile.

        Array assumes that the pressure is in Pa.

        Parameters
        ----------
        array : npt.NDArray[np.float64]
            Pressure profile array

        reverse : t.Optional[bool], optional
            Reverse the pressure profile, by default False

        is_central : t.Optional[bool], optional
            If the pressure profile is central or not, by default True
            **New in v3.2.0**

        """
        super().__init__(self.__class__.__name__, array.shape[-1])

        self.is_central = is_central

        if reverse:
            self.pressure_array = array[::-1]
        else:
            self.pressure_array = array

    def compute_pressure_profile(self) -> None:
        """Sets up the pressure profile for the atmosphere model."""
        if self.is_central:
            logp = np.log10(self.pressure_profile)
            gradp = np.gradient(logp)

            self.pressure_profile_levels = 10 ** np.append(
                logp - gradp / 2, logp[-1] + gradp[-1] / 2
            )
        else:
            self.pressure_profile_levels = self.pressure_profile

            self.pressure_profile = self.pressure_profile_levels[:-1] * np.sqrt(
                self.pressure_profile_levels[1:] / self.pressure_profile_levels[:-1]
            )

    @property
    def profile(self) -> npt.NDArray[np.float64]:
        """Pressure profile array."""
        return self.pressure_profile

    def write(self, output: OutputGroup) -> OutputGroup:
        """Write pressure profile to output."""
        pressure = super().write(output)

        return pressure

    @classmethod
    def input_keywords(cls) -> t.Tuple[str, ...]:
        """Input keywords for pressure profile."""
        return (
            "array",
            "fromarray",
        )



"""Base stellar class."""
import typing as t

import numpy as np
import numpy.typing as npt

from taurex.constants import MSOL, RSOL
from taurex.data.fittable import Fittable
from taurex.log import Logger
from taurex.output import OutputGroup, Writeable
from taurex.util.emission import black_body

from ..citation import Citable


class Star(Fittable, Logger, Writeable, Citable):
    """A base class that holds information on the star in the model.
    Its implementation is a star that has a blackbody spectrum.

    """

    def __init__(
        self,
        temperature: t.Optional[float] = 5000,
        radius: t.Optional[float] = 1.0,
        distance: t.Optional[float] = 1,
        magnitudeK: t.Optional[float] = 10.0,  # noqa: N803
        mass: t.Optional[float] = 1.0,
        metallicity: t.Optional[float] = 1.0,
    ):
        """Initialize a star.

        Parameters
        ----------

        temperature: float, optional
            Stellar temperature in Kelvin

        radius: float, optional
            Stellar radius in Solar radius

        metallicity: float, optional
            Metallicity in solar values

        mass: float, optional
            Stellar mass in solar mass

        distance: float, optional
            Distance from Earth in pc

        magnitudeK: float, optional
            Maginitude in K band

        """
        Logger.__init__(self, self.__class__.__name__)
        Fittable.__init__(self)
        self._temperature = temperature
        self._radius = radius * RSOL
        self._mass = mass * MSOL
        self.debug("Star mass %s", self._mass)
        self.sed = None
        self.distance = distance
        self.magnitudeK = magnitudeK
        self._metallicity = metallicity

    @property
    def radius(self) -> float:
        """Radius in metres."""
        return self._radius

    @property
    def temperature(self) -> float:
        """Blackbody temperature in Kelvin."""
        return self._temperature

    @temperature.setter
    def temperature(self, value: float) -> None:
        """Set blackbody temperature in Kelvin."""
        self._temperature = value

    @property
    def mass(self) -> float:
        """Mass in kg."""
        return self._mass

    def initialize(self, wngrid: npt.NDArray[np.float64]) -> None:
        """Initializes the blackbody spectrum on the given wavenumber grid

        Parameters
        ----------
        wngrid: :obj:`array`
            Wavenumber grid cm-1 to compute black body spectrum

        """
        self.sed = black_body(wngrid, self.temperature)

    @property
    def spectralEmissionDensity(self) -> npt.NDArray[np.float64]:  # noqa: N802
        """Spectral emmision density in W/m2/cm-1/sr."""
        return self.sed

    def write(self, output: OutputGroup) -> OutputGroup:
        """Write to output group."""
        star = output.create_group("Star")
        star.write_string("star_type", self.__class__.__name__)
        star.write_scalar("temperature", self.temperature)
        star.write_scalar("radius", self._radius / RSOL)
        star.write_scalar("distance", self.distance)
        star.write_scalar("mass", self._mass / MSOL)
        star.write_scalar("magnitudeK", self.magnitudeK)
        star.write_scalar("metallicity", self._metallicity)
        star.write_scalar("radius_m", self.radius)
        star.write_array("SED", self.spectralEmissionDensity)
        star.write_scalar("mass_kg", self._mass)
        return star

    @classmethod
    def input_keywords(cls) -> t.Tuple[str, ...]:
        """Input keywords for star."""
        raise NotImplementedError


class BlackbodyStar(Star):
    """Alias for the base star type"""

    @classmethod
    def input_keywords(cls) -> t.Tuple[str, ...]:
        return ("blackbody",)



"""Ktables from pickle files."""
import pathlib
import pickle  # noqa: S403
import typing as t

import numpy as np
import numpy.typing as npt

from taurex.types import PathLike

from ..interpolateopacity import InterpModeType, InterpolatingOpacity
from .ktable import KTable


class PickleKTable(KTable, InterpolatingOpacity):
    """Ktables loaded from pickle files.

    This is the old way of loading ktables from pickle files
    from TauREx 1/2. It is not recommended to use this format
    of loading ktables as it is not as flexible as the HDF5
    format. It is recommended to use the HDF5 format instead.

    """

    @classmethod
    def discover(cls) -> t.List[t.Tuple[str, t.Tuple[pathlib.Path, InterpModeType]]]:
        import pathlib

        from taurex.cache import GlobalCache
        from taurex.util import sanitize_molecule_string

        path = GlobalCache()["ktable_path"]
        if path is None:
            return []
        path = pathlib.Path(path)

        files = path.glob("*.pickle")

        discovery = []

        interp = GlobalCache()["xsec_interpolation"] or "linear"

        for f in files:
            splits = pathlib.Path(f).stem.split(".")
            mol_name = sanitize_molecule_string(splits[0])

            discovery.append((mol_name, [f, interp]))

        return discovery

    def __init__(
        self, filename: PathLike, interpolation_mode: t.Optional[PathLike] = "linear"
    ) -> None:
        """Initialize PickleKTable.

        Parameters
        ----------
        filename : PathLike
            Path to file.
        interpolation_mode : t.Optional[PathLike], optional
            Interpolation mode, by default "linear".


        """
        filename = pathlib.Path(filename)
        super().__init__(
            f"PickleKtable:{filename.stem[0:10]}",
            interpolation_mode=interpolation_mode,
        )
        if not filename.exists():
            raise FileNotFoundError(f"File {filename} not found")

        self._filename = filename
        self._molecule_name = None
        self._spec_dict = None
        self._resolution = None
        self._load_pickle_file(filename)

    @property
    def moleculeName(self) -> str:  # noqa: N802
        """Molecule name."""
        return self._molecule_name

    @property
    def xsecGrid(self) -> npt.NDArray[np.float64]:  # noqa: N802
        """Cross section grid."""
        return self._xsec_grid

    def _load_pickle_file(self, filename: pathlib.Path) -> None:
        """Load pickle file."""
        self.info(f"Loading opacity from {filename}")
        try:
            with open(filename, "rb") as f:
                self._spec_dict = pickle.load(f)  # noqa: S301
        except UnicodeDecodeError:
            with open(filename, "rb") as f:
                self._spec_dict = pickle.load(f, encoding="latin1")  # noqa: S301

        self._wavenumber_grid = self._spec_dict["bin_centers"]
        self._ngauss = self._spec_dict["ngauss"]
        self._temperature_grid = self._spec_dict["t"]
        self._pressure_grid = self._spec_dict["p"] * 1e5
        self._xsec_grid = self._spec_dict["kcoeff"]
        self._weights = self._spec_dict["weights"]
        self._molecule_name = self._spec_dict["name"]

        self._min_pressure = self._pressure_grid.min()
        self._max_pressure = self._pressure_grid.max()
        self._min_temperature = self._temperature_grid.min()
        self._max_temperature = self._temperature_grid.max()
        self.clean_molecule_name()

    def clean_molecule_name(self) -> None:
        """Clean molecule name."""
        splits = self.moleculeName.split("_")
        self._molecule_name = splits[0]

    @property
    def wavenumberGrid(self) -> npt.NDArray[np.float64]:  # noqa: N802
        """Wavenumber grid."""
        return self._wavenumber_grid

    @property
    def temperatureGrid(self) -> npt.NDArray[np.float64]:  # noqa: N802
        """Temperature grid."""
        return self._temperature_grid

    @property
    def pressureGrid(self) -> npt.NDArray[np.float64]:  # noqa: N802
        """Pressure grid."""
        return self._pressure_grid

    @property
    def resolution(self) -> float:
        """Resolution."""
        return self._resolution

    @property
    def weights(self) -> npt.NDArray[np.float64]:
        """Quadrature Weights."""
        return self._weights
